# pypl2lib.py - Classes and functions for accessing functions
# in PL2FileReader.dll
#
# (c) 2016 Plexon, Inc., Dallas, Texas
# www.plexon.com
#
# This software is provided as-is, without any warranty.
# You are free to modify or share this file, provided that the above
# copyright notice is kept intact.

import platform
import subprocess
import pathlib
import warnings
import numpy as np

platform_is_windows = platform.system() == "Windows"

if platform_is_windows:
    import ctypes
else:
    is_wine_available = subprocess.run(["which", "wine"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=False)
    if is_wine_available.returncode != 0:
        raise ImportError("Wine is not installed. Please install wine to use the PL2FileReader.dll")

    from zugbruecke import CtypesSession

    if platform.system() == "Darwin":
        ctypes = CtypesSession(log_level=100, arch="win64")
    else:
        ctypes = CtypesSession(log_level=100)


class tm(ctypes.Structure):
    _fields_ = [
        ("tm_sec", ctypes.c_int),
        ("tm_min", ctypes.c_int),
        ("tm_hour", ctypes.c_int),
        ("tm_mday", ctypes.c_int),
        ("tm_mon", ctypes.c_int),
        ("tm_year", ctypes.c_int),
        ("tm_wday", ctypes.c_int),
        ("tm_yday", ctypes.c_int),
        ("tm_isdst", ctypes.c_int),
    ]


class PL2FileInfo(ctypes.Structure):
    _fields_ = [
        ("m_CreatorComment", ctypes.c_char * 256),
        ("m_CreatorSoftwareName", ctypes.c_char * 64),
        ("m_CreatorSoftwareVersion", ctypes.c_char * 16),
        ("m_CreatorDateTime", tm),
        ("m_CreatorDateTimeMilliseconds", ctypes.c_int),
        ("m_TimestampFrequency", ctypes.c_double),
        ("m_NumberOfChannelHeaders", ctypes.c_uint),
        ("m_TotalNumberOfSpikeChannels", ctypes.c_uint),
        ("m_NumberOfRecordedSpikeChannels", ctypes.c_uint),
        ("m_TotalNumberOfAnalogChannels", ctypes.c_uint),
        ("m_NumberOFRecordedAnalogChannels", ctypes.c_uint),
        ("m_NumberOfDigitalChannels", ctypes.c_uint),
        ("m_MinimumTrodality", ctypes.c_uint),
        ("m_MaximumTrodality", ctypes.c_uint),
        ("m_NumberOfNonOmniPlexSources", ctypes.c_uint),
        ("m_Unused", ctypes.c_int),
        ("m_ReprocessorComment", ctypes.c_char * 256),
        ("m_ReprocessorSoftwareName", ctypes.c_char * 64),
        ("m_ReprocessorSoftwareVersion", ctypes.c_char * 16),
        ("m_ReprocessorDateTime", tm),
        ("m_ReprocessorDateTimeMilliseconds", ctypes.c_int),
        ("m_StartRecordingTime", ctypes.c_ulonglong),
        ("m_DurationOfRecording", ctypes.c_ulonglong),
    ]


class PL2AnalogChannelInfo(ctypes.Structure):
    _fields_ = [
        ("m_Name", ctypes.c_char * 64),
        ("m_Source", ctypes.c_uint),
        ("m_Channel", ctypes.c_uint),
        ("m_ChannelEnabled", ctypes.c_uint),
        ("m_ChannelRecordingEnabled", ctypes.c_uint),
        ("m_Units", ctypes.c_char * 16),
        ("m_SamplesPerSecond", ctypes.c_double),
        ("m_CoeffToConvertToUnits", ctypes.c_double),
        ("m_SourceTrodality", ctypes.c_uint),
        ("m_OneBasedTrode", ctypes.c_ushort),
        ("m_OneBasedChannelInTrode", ctypes.c_ushort),
        ("m_NumberOfValues", ctypes.c_ulonglong),
        ("m_MaximumNumberOfFragments", ctypes.c_ulonglong),
    ]


class PL2SpikeChannelInfo(ctypes.Structure):
    _fields_ = [
        ("m_Name", ctypes.c_char * 64),
        ("m_Source", ctypes.c_uint),
        ("m_Channel", ctypes.c_uint),
        ("m_ChannelEnabled", ctypes.c_uint),
        ("m_ChannelRecordingEnabled", ctypes.c_uint),
        ("m_Units", ctypes.c_char * 16),
        ("m_SamplesPerSecond", ctypes.c_double),
        ("m_CoeffToConvertToUnits", ctypes.c_double),
        ("m_SamplesPerSpike", ctypes.c_uint),
        ("m_Threshold", ctypes.c_int),
        ("m_PreThresholdSamples", ctypes.c_uint),
        ("m_SortEnabled", ctypes.c_uint),
        ("m_SortMethod", ctypes.c_uint),
        ("m_NumberOfUnits", ctypes.c_uint),
        ("m_SortRangeStart", ctypes.c_uint),
        ("m_SortRangeEnd", ctypes.c_uint),
        ("m_UnitCounts", ctypes.c_ulonglong * 256),
        ("m_SourceTrodality", ctypes.c_uint),
        ("m_OneBasedTrode", ctypes.c_ushort),
        ("m_OneBasedChannelInTrode", ctypes.c_ushort),
        ("m_NumberOfSpikes", ctypes.c_ulonglong),
    ]


class PL2DigitalChannelInfo(ctypes.Structure):
    _fields_ = [
        ("m_Name", ctypes.c_char * 64),
        ("m_Source", ctypes.c_uint),
        ("m_Channel", ctypes.c_uint),
        ("m_ChannelEnabled", ctypes.c_uint),
        ("m_ChannelRecordingEnabled", ctypes.c_uint),
        ("m_NumberOfEvents", ctypes.c_ulonglong),
    ]


def to_array(c_array):
    return np.ctypeslib.as_array(c_array)


def to_array_nonzero(c_array):
    a = np.ctypeslib.as_array(c_array)
    return a[np.where(a)]


class PyPL2FileReader:
    def __init__(self, pl2_dll_file_path=None):
        """
        PyPL2FileReader class implements functions in the C++ PL2 File Reader
        API provided by Plexon, Inc.

        Args:
            pl2_dll_file_path - path where PL2FileReader.dll is location.
                The default value assumes the .dll files are located in the
                'bin' directory, which is a subdirectory of this package.
                Any file path passed is converted to an absolute path and checked
                to see if the .dll exists there.

        Returns:
            None
        """
        self._file_handle = ctypes.c_int(0)
        self.pl2_file_info = None
        if pl2_dll_file_path is None:
            pl2_dll_file_path = pathlib.Path(__file__).parents[1] / "bin" / "PL2FileReader.dll"
        self.pl2_dll_path = pathlib.Path(pl2_dll_file_path).absolute()

        # use default '32bit' dll version
        self.pl2_dll_file_path = self.pl2_dll_path

        try:
            self.pl2_dll = ctypes.CDLL(str(self.pl2_dll_file_path))
        except IOError:
            raise IOError(
                f"Error: Can't load PL2FileReader.dll at: {self.pl2_dll_file_path}. "
                "PL2FileReader.dll is bundled with the C++ PL2 Offline Files SDK "
                "located on the Plexon Inc website: www.plexon.com "
                "Contact Plexon Support for more information: support@plexon.com"
            )

    def pl2_open_file(self, pl2_file):
        """
        Opens and returns a handle to a PL2 file.

        Args:
            pl2_file - full path of the file

        Returns:
            None

        """
        if isinstance(pl2_file, pathlib.Path):
            pl2_file = str(pl2_file)
        self.pl2_dll.PL2_OpenFile.argtypes = (
            ctypes.POINTER(ctypes.c_char),
            ctypes.POINTER(ctypes.c_int),
        )
        self.pl2_dll.PL2_OpenFile.memsync = [
            {
                "p": [0],  # ctypes.POINTER argument
                "n": True,  # null-terminated string flag
            }
        ]

        self.pl2_dll.PL2_OpenFile(
            pl2_file.encode("ascii"),
            ctypes.byref(self._file_handle),
        )

        # load file info
        self.pl2_get_file_info()
        # check if spiking data can be loaded using zugbruecke
        self._check_spike_channel_data_consistency()

    def pl2_close_file(self):
        """
        Closes handle to PL2 file.

        Returns:
            None
        """

        self.pl2_dll.PL2_CloseFile.argtypes = (ctypes.POINTER(ctypes.c_int),)
        self.pl2_dll.PL2_CloseFile(ctypes.c_int(1))

    def pl2_close_all_files(self):
        """
        Closes all files that have been opened by the .dll

        Args:
            None

        Returns:
            None
        """

        self.pl2_dll.PL2_CloseAllFiles.argtypes = ()
        self.pl2_dll.PL2_CloseAllFiles()

    def pl2_get_last_error(self):
        """
        Retrieve description of the last error

        Returns:
            str - error message
        """

        self.pl2_dll.PL2_GetLastError.argtypes = (
            ctypes.POINTER(ctypes.c_char),
            ctypes.c_int,
        )
        self.pl2_dll.PL2_GetLastError.memsync = [{"p": [0], "l": [1], "t": ctypes.c_char}]

        buffer = (ctypes.c_char * 256)()
        self.pl2_dll.PL2_GetLastError(buffer, ctypes.c_int(256))

        return str(buffer.value)

    def _check_spike_channel_data_consistency(self):
        """
        Check if all spiking channels use the same number of samples per
        waveform. Only in this case can zugbruecke reliably load spiking data
        """

        if not self.pl2_file_info.m_TotalNumberOfSpikeChannels:
            return

        # extract samples per spike of first channel
        channel_info = self.pl2_get_spike_channel_info(0)
        n_samples_per_spike = channel_info.m_SamplesPerSpike

        # compare with all other channels
        for i in range(1, self.pl2_file_info.m_TotalNumberOfSpikeChannels):
            channel_info = self.pl2_get_spike_channel_info(i)

            if channel_info.m_SamplesPerSpike != n_samples_per_spike:
                warnings.warn(
                    "The spike channels contain different number of samples per spike. "
                    "Spiking data can probably not be loaded using zugbruecke. "
                    "Use a windows operating system or remove the offending channels "
                    "from the file."
                )
                return

    def pl2_get_file_info(self):
        """
        Retrieve information about pl2 file.

        Returns:
            pl2_file_info - PL2FileInfo class instance
        """

        self.pl2_file_info = PL2FileInfo()

        self.pl2_dll.PL2_GetFileInfo.argtypes = (
            ctypes.c_int,
            ctypes.POINTER(PL2FileInfo),
        )

        result = self.pl2_dll.PL2_GetFileInfo(self._file_handle, ctypes.byref(self.pl2_file_info))

        # If res is 0, print error message
        if result == 0:
            self._print_error()
            return None

        return self.pl2_file_info

    def pl2_get_analog_channel_info(self, zero_based_channel_index):
        """
        Retrieve information about an analog channel

        Args:
            zero_based_channel_index - zero-based analog channel index

        Returns:
            pl2_analog_channel_info - PL2AnalogChannelInfo class instance
        """

        self.pl2_dll.PL2_GetAnalogChannelInfo.argtypes = (
            ctypes.c_int,
            ctypes.c_int,
            ctypes.POINTER(PL2AnalogChannelInfo),
        )

        pl2_analog_channel_info = PL2AnalogChannelInfo()
        result = self.pl2_dll.PL2_GetAnalogChannelInfo(
            self._file_handle, ctypes.c_int(zero_based_channel_index), ctypes.byref(pl2_analog_channel_info)
        )

        if not result:
            self._print_error()
            return None

        return pl2_analog_channel_info

    def pl2_get_analog_channel_info_by_name(self, channel_name):
        """
        Retrieve information about an analog channel

        Args:
            channel_name - analog channel name

        Returns:
            pl2_analog_channel_info - PL2AnalogChannelInfo class instance
        """

        self.pl2_dll.PL2_GetAnalogChannelInfoByName.argtypes = (
            ctypes.c_int,
            ctypes.POINTER(ctypes.c_char),
            ctypes.POINTER(PL2AnalogChannelInfo),
        )

        self.pl2_dll.PL2_GetAnalogChannelInfoByName.memsync = [
            {
                "p": [1],
                "n": True,  # null-terminated string flag
            }
        ]

        if hasattr(channel_name, "encode"):
            channel_name = channel_name.encode("ascii")

        pl2_analog_channel_info = PL2AnalogChannelInfo()

        result = self.pl2_dll.PL2_GetAnalogChannelInfoByName(
            self._file_handle, channel_name, ctypes.byref(pl2_analog_channel_info)
        )

        if not result:
            self._print_error()
            return None
        return pl2_analog_channel_info

    def pl2_get_analog_channel_info_by_source(self, source_id, one_based_channel_index_in_source):
        """
        Retrieve information about an analog channel

        Args:
            source_id - numeric source ID
            one_based_channel_index_in_source - one-based channel index within the source

        Returns:
            pl2_analog_channel_info - PL2AnalogChannelInfo class instance
        """

        self.pl2_dll.PL2_GetAnalogChannelInfoBySource.argtypes = (
            ctypes.c_int,
            ctypes.c_int,
            ctypes.c_int,
            ctypes.POINTER(PL2AnalogChannelInfo),
        )

        pl2_analog_channel_info = PL2AnalogChannelInfo()
        result = self.pl2_dll.PL2_GetAnalogChannelInfoBySource(
            self._file_handle,
            ctypes.c_int(source_id),
            ctypes.c_int(one_based_channel_index_in_source),
            ctypes.byref(pl2_analog_channel_info),
        )

        if not result:
            self._print_error()
            return None

        return pl2_analog_channel_info

    def pl2_get_analog_channel_data(self, zero_based_channel_index):
        """
        Retrieve analog channel data

        Args:
            zero_based_channel_index - zero based channel index

        Returns:
            fragment_timestamps - array the size of PL2AnalogChannelInfo.m_MaximumNumberOfFragments
            fragment_counts - array the size of PL2AnalogChannelInfo.m_MaximumNumberOfFragments
            values - array the size of PL2AnalogChannelInfo.m_NumberOfValues
        """

        achannel_info = self.pl2_get_analog_channel_info(zero_based_channel_index)

        num_fragments_returned = ctypes.c_ulonglong(achannel_info.m_MaximumNumberOfFragments)
        num_data_points_returned = ctypes.c_ulonglong(achannel_info.m_NumberOfValues)
        fragment_timestamps = (ctypes.c_longlong * achannel_info.m_MaximumNumberOfFragments)()
        fragment_counts = (ctypes.c_ulonglong * achannel_info.m_MaximumNumberOfFragments)()
        values = (ctypes.c_short * achannel_info.m_NumberOfValues)()

        self.pl2_dll.PL2_GetAnalogChannelData.argtypes = (
            ctypes.c_int,
            ctypes.c_int,
            ctypes.POINTER(ctypes.c_ulonglong),
            ctypes.POINTER(ctypes.c_ulonglong),
            ctypes.POINTER(ctypes.c_longlong),
            ctypes.POINTER(ctypes.c_ulonglong),
            ctypes.POINTER(ctypes.c_short),
        )

        self.pl2_dll.PL2_GetAnalogChannelData.memsync = [
            {"p": [4], "l": [2], "t": ctypes.c_longlong},
            {"p": [5], "l": [2], "t": ctypes.c_ulonglong},
            {"p": [6], "l": [3], "t": ctypes.c_short},
        ]

        result = self.pl2_dll.PL2_GetAnalogChannelData(
            self._file_handle,
            ctypes.c_int(zero_based_channel_index),
            num_fragments_returned,
            num_data_points_returned,
            fragment_timestamps,
            fragment_counts,
            values,
        )

        if not result:
            self._print_error()
            return None

        return to_array(fragment_timestamps), to_array(fragment_counts), to_array(values)

    def pl2_get_analog_channel_data_subset(self):
        pass

    def pl2_get_analog_channel_data_by_name(self, channel_name):
        """
        Retrieve analog channel data

        Args:
            channel_name - analog channel name

        Returns:
            fragment_timestamps - array the size of PL2AnalogChannelInfo.m_MaximumNumberOfFragments
            fragment_counts - array the size of PL2AnalogChannelInfo.m_MaximumNumberOfFragments
            values - array the size of PL2AnalogChannelInfo.m_NumberOfValues
        """

        if hasattr(channel_name, "encode"):
            channel_name = channel_name.encode("ascii")

        achannel_info = self.pl2_get_analog_channel_info_by_name(channel_name)

        num_fragments_returned = ctypes.c_ulonglong(achannel_info.m_MaximumNumberOfFragments)
        num_data_points_returned = ctypes.c_ulonglong(achannel_info.m_NumberOfValues)
        fragment_timestamps = (ctypes.c_longlong * achannel_info.m_MaximumNumberOfFragments)()
        fragment_counts = (ctypes.c_ulonglong * achannel_info.m_MaximumNumberOfFragments)()
        values = (ctypes.c_short * achannel_info.m_NumberOfValues)()

        self.pl2_dll.PL2_GetAnalogChannelDataByName.argtypes = (
            ctypes.c_int,
            ctypes.POINTER(ctypes.c_char),
            ctypes.POINTER(ctypes.c_ulonglong),
            ctypes.POINTER(ctypes.c_ulonglong),
            ctypes.POINTER(ctypes.c_longlong),
            ctypes.POINTER(ctypes.c_ulonglong),
            ctypes.POINTER(ctypes.c_short),
        )

        self.pl2_dll.PL2_GetAnalogChannelDataByName.memsync = [
            {
                "p": [1],
                "n": True,  # null-terminated string flag
            },
            {"p": [4], "l": [2], "t": ctypes.c_longlong},
            {"p": [5], "l": [2], "t": ctypes.c_ulonglong},
            {"p": [6], "l": [3], "t": ctypes.c_short},
        ]

        result = self.pl2_dll.PL2_GetAnalogChannelDataByName(
            self._file_handle,
            channel_name,
            num_fragments_returned,
            num_data_points_returned,
            fragment_timestamps,
            fragment_counts,
            values,
        )

        if not result:
            self._print_error()
            return None

        return to_array(fragment_timestamps), to_array(fragment_counts), to_array(values)

    def pl2_get_analog_channel_data_by_source(self, source_id, one_based_channel_index_in_source):
        """
        Retrieve analog channel data

        Args:
            source_id - numeric source ID
            one_based_channel_index_in_source - one-based channel index within the source

        Returns:
            fragment_timestamps - array the size of PL2AnalogChannelInfo.m_MaximumNumberOfFragments
            fragment_counts - array the size of PL2AnalogChannelInfo.m_MaximumNumberOfFragments
            values - array the size of PL2AnalogChannelInfo.m_NumberOfValues
        """

        achannel_info = self.pl2_get_analog_channel_info_by_source(source_id, one_based_channel_index_in_source)

        num_fragments_returned = ctypes.c_ulonglong(achannel_info.m_MaximumNumberOfFragments)
        num_data_points_returned = ctypes.c_ulonglong(achannel_info.m_NumberOfValues)
        fragment_timestamps = (ctypes.c_longlong * achannel_info.m_MaximumNumberOfFragments)()
        fragment_counts = (ctypes.c_ulonglong * achannel_info.m_MaximumNumberOfFragments)()
        values = (ctypes.c_short * achannel_info.m_NumberOfValues)()

        self.pl2_dll.PL2_GetAnalogChannelDataBySource.argtypes = (
            ctypes.c_int,
            ctypes.c_int,
            ctypes.c_int,
            ctypes.POINTER(ctypes.c_ulonglong),
            ctypes.POINTER(ctypes.c_ulonglong),
            ctypes.POINTER(ctypes.c_longlong),
            ctypes.POINTER(ctypes.c_ulonglong),
            ctypes.POINTER(ctypes.c_short),
        )

        self.pl2_dll.PL2_GetAnalogChannelDataBySource.memsync = [
            {"p": [5], "l": [3], "t": ctypes.c_longlong},
            {"p": [6], "l": [3], "t": ctypes.c_ulonglong},
            {"p": [7], "l": [4], "t": ctypes.c_short},
        ]

        result = self.pl2_dll.PL2_GetAnalogChannelDataBySource(
            self._file_handle,
            ctypes.c_int(source_id),
            ctypes.c_int(one_based_channel_index_in_source),
            num_fragments_returned,
            num_data_points_returned,
            fragment_timestamps,
            fragment_counts,
            values,
        )

        if not result:
            self._print_error()
            return None

        return to_array(fragment_timestamps), to_array(fragment_counts), to_array(values)

    def pl2_get_spike_channel_info(self, zero_based_channel_index):
        """
        Retrieve information about a spike channel

        Args:
            zero_based_channel_index - zero-based spike channel index

        Returns:
            pl2_spike_channel_info - PL2SpikeChannelInfo class instance
        """

        pl2_spike_channel_info = PL2SpikeChannelInfo()

        self.pl2_dll.PL2_GetSpikeChannelInfo.argtypes = (
            ctypes.c_int,
            ctypes.c_int,
            ctypes.POINTER(PL2SpikeChannelInfo),
        )

        result = self.pl2_dll.PL2_GetSpikeChannelInfo(
            self._file_handle, ctypes.c_int(zero_based_channel_index), ctypes.byref(pl2_spike_channel_info)
        )

        if not result:
            self._print_error()
            return None

        return pl2_spike_channel_info

    def pl2_get_spike_channel_info_by_name(self, channel_name):
        """
        Retrieve information about a spike channel

        Args:
            channel_name - spike channel name

        Returns:
            pl2_spike_channel_info - PL2SpikeChannelInfo class instance
        """

        self.pl2_dll.PL2_GetSpikeChannelInfoByName.argtypes = (
            ctypes.c_int,
            ctypes.POINTER(ctypes.c_char),
            ctypes.POINTER(PL2SpikeChannelInfo),
        )

        if hasattr(channel_name, "encode"):
            channel_name = channel_name.encode("ascii")

        self.pl2_dll.PL2_GetSpikeChannelInfoByName.memsync = [
            {
                "p": [1],
                "n": True,  # null-terminated string flag
            }
        ]

        pl2_spike_channel_info = PL2SpikeChannelInfo()

        result = self.pl2_dll.PL2_GetSpikeChannelInfoByName(
            self._file_handle, channel_name, ctypes.byref(pl2_spike_channel_info)
        )
        if not result:
            self._print_error()
            return None
        return pl2_spike_channel_info

    def pl2_get_spike_channel_info_by_source(self, source_id, one_based_channel_index_in_source):
        """
        Retrieve information about a spike channel

        Args:
            source_id - numeric source ID
            one_based_channel_index_in_source - one-based channel index within the source

        Returns:
            pl2_spike_channel_info - PL2SpikeChannelInfo class instance
        """

        pl2_spike_channel_info = PL2SpikeChannelInfo()

        self.pl2_dll.PL2_GetSpikeChannelInfoBySource.argtypes = (
            ctypes.c_int,
            ctypes.c_int,
            ctypes.c_int,
            ctypes.POINTER(PL2SpikeChannelInfo),
        )
        result = self.pl2_dll.PL2_GetSpikeChannelInfoBySource(
            self._file_handle,
            ctypes.c_int(source_id),
            ctypes.c_int(one_based_channel_index_in_source),
            ctypes.byref(pl2_spike_channel_info),
        )

        if not result:
            self._print_error()
            return None

        return pl2_spike_channel_info

    def pl2_get_spike_channel_data(self, zero_based_channel_index):
        """
        Retrieve spike channel data

        Args:
            zero_based_channel_index - zero based channel index

        Returns:
            spike_timestamps - array the size of PL2SpikeChannelInfo.m_NumberOfSpikes
            units - array the size of PL2SpikeChannelInfo.m_NumberOfSpikes
            values - array the size of (PL2SpikeChannelInfo.m_NumberOfSpikes * PL2SpikeChannelInfo.m_SamplesPerSpike)
        """

        # extracting m_SamplesPerSpike to prepare data reading
        # This solution only works if all channels have the same number of samples per spike
        # as ctypes / zugbruecke is caching the memsync attribute once defined once
        schannel_info = self.pl2_get_spike_channel_info(zero_based_channel_index)
        samples_per_spike = schannel_info.m_SamplesPerSpike

        self.pl2_dll.PL2_GetSpikeChannelData.argtypes = (
            ctypes.c_int,
            ctypes.c_int,
            ctypes.POINTER(ctypes.c_ulonglong),
            ctypes.POINTER(ctypes.c_ulonglong),
            ctypes.POINTER(ctypes.c_ushort),
            ctypes.POINTER(ctypes.c_short),
        )

        self.pl2_dll.PL2_GetSpikeChannelData.memsync = [
            {"p": [3], "l": [2], "t": ctypes.c_longlong},
            {"p": [4], "l": [2], "t": ctypes.c_ushort},
            {"p": [5], "l": ([2],), "func": f"lambda x: x.value * {samples_per_spike}", "t": ctypes.c_short},
        ]

        # These will be filled in by the dll method.
        num_spikes_returned = ctypes.c_ulonglong(schannel_info.m_NumberOfSpikes)
        spike_timestamps = (ctypes.c_ulonglong * schannel_info.m_NumberOfSpikes)()
        units = (ctypes.c_ushort * schannel_info.m_NumberOfSpikes)()
        values = (ctypes.c_short * (schannel_info.m_NumberOfSpikes * schannel_info.m_SamplesPerSpike))()

        result = self.pl2_dll.PL2_GetSpikeChannelData(
            self._file_handle,
            ctypes.c_int(zero_based_channel_index),
            num_spikes_returned,
            spike_timestamps,
            units,
            values,
        )

        if not result:
            self._print_error()
            return None

        spike_array_shape = (schannel_info.m_NumberOfSpikes, schannel_info.m_SamplesPerSpike)

        return to_array(spike_timestamps), to_array(units), to_array(values).reshape(spike_array_shape)

    def pl2_get_spike_channel_data_by_name(self, channel_name):
        """
        Retrieve spike channel data

        Args:
            channel_name = channel name

        Returns:
            spike_timestamps - array the size of PL2SpikeChannelInfo.m_NumberOfSpikes
            units - array the size of PL2SpikeChannelInfo.m_NumberOfSpikes
            values - array the size of (PL2SpikeChannelInfo.m_NumberOfSpikes * PL2SpikeChannelInfo.m_SamplesPerSpike)
        """

        if hasattr(channel_name, "encode"):
            channel_name = channel_name.encode("ascii")

        self.pl2_dll.PL2_GetSpikeChannelDataByName.argtypes = (
            ctypes.c_int,
            ctypes.POINTER(ctypes.c_char),
            ctypes.POINTER(ctypes.c_ulonglong),
            ctypes.POINTER(ctypes.c_ulonglong),
            ctypes.POINTER(ctypes.c_ushort),
            ctypes.POINTER(ctypes.c_short),
        )

        # extracting m_SamplesPerSpike to prepare data reading
        # This solution only works if all channels have the same number of samples per spike
        # as ctypes / zugbruecke is caching the memsync attribute once defined once
        schannel_info = self.pl2_get_spike_channel_info_by_name(channel_name)
        samples_per_spike = schannel_info.m_SamplesPerSpike

        self.pl2_dll.PL2_GetSpikeChannelDataByName.memsync = [
            {
                "p": [1],
                "n": True,  # null-terminated string flag
            },
            {"p": [3], "l": [2], "t": ctypes.c_longlong},
            {"p": [4], "l": [2], "t": ctypes.c_ushort},
            {"p": [5], "l": ([2],), "func": f"lambda x: x.value * {samples_per_spike}", "t": ctypes.c_short},
        ]

        # These will be filled in by the dll method.
        num_spikes_returned = ctypes.c_ulonglong(schannel_info.m_NumberOfSpikes)
        spike_timestamps = (ctypes.c_ulonglong * schannel_info.m_NumberOfSpikes)()
        units = (ctypes.c_ushort * schannel_info.m_NumberOfSpikes)()
        values = (ctypes.c_short * (schannel_info.m_NumberOfSpikes * schannel_info.m_SamplesPerSpike))()

        result = self.pl2_dll.PL2_GetSpikeChannelDataByName(
            self._file_handle, channel_name, num_spikes_returned, spike_timestamps, units, values
        )

        if not result:
            self._print_error()
            return None

        spike_array_shape = (schannel_info.m_NumberOfSpikes, schannel_info.m_SamplesPerSpike)

        return to_array(spike_timestamps), to_array(units), to_array(values).reshape(spike_array_shape)

    def pl2_get_spike_channel_data_by_source(self, source_id, one_based_channel_index_in_source):
        """
        Retrieve spike channel data

        Args:
            source_id - numeric source ID
            one_based_channel_index_in_source - one-based channel index within the source

        Returns:
            spike_timestamps - array the size of PL2SpikeChannelInfo.m_NumberOfSpikes
            units - array the size of PL2SpikeChannelInfo.m_NumberOfSpikes
            values - array the size of (PL2SpikeChannelInfo.m_NumberOfSpikes * PL2SpikeChannelInfo.m_SamplesPerSpike)
        """

        self.pl2_dll.PL2_GetSpikeChannelDataBySource.argtypes = (
            ctypes.c_int,
            ctypes.c_int,
            ctypes.c_int,
            ctypes.POINTER(ctypes.c_ulonglong),
            ctypes.POINTER(ctypes.c_ulonglong),
            ctypes.POINTER(ctypes.c_ushort),
            ctypes.POINTER(ctypes.c_short),
        )

        # extracting m_SamplesPerSpike to prepare data reading
        # This solution only works if all channels have the same number of samples per spike
        # as ctypes / zugbruecke is caching the memsync attribute once defined once
        schannel_info = self.pl2_get_spike_channel_info_by_source(source_id, one_based_channel_index_in_source)
        samples_per_spike = schannel_info.m_SamplesPerSpike

        self.pl2_dll.PL2_GetSpikeChannelDataBySource.memsync = [
            {"p": [4], "l": [3], "t": ctypes.c_longlong},
            {"p": [5], "l": [3], "t": ctypes.c_ushort},
            {"p": [6], "l": ([3],), "func": f"lambda x: x.value * {samples_per_spike}", "t": ctypes.c_short},
        ]

        # These will be filled in by the dll method.
        num_spikes_returned = ctypes.c_ulonglong(schannel_info.m_NumberOfSpikes)
        spike_timestamps = (ctypes.c_ulonglong * schannel_info.m_NumberOfSpikes)()
        units = (ctypes.c_ushort * schannel_info.m_NumberOfSpikes)()
        values = (ctypes.c_short * (schannel_info.m_NumberOfSpikes * schannel_info.m_SamplesPerSpike))()

        result = self.pl2_dll.PL2_GetSpikeChannelDataBySource(
            self._file_handle,
            ctypes.c_int(source_id),
            ctypes.c_int(one_based_channel_index_in_source),
            num_spikes_returned,
            spike_timestamps,
            units,
            values,
        )

        if not result:
            self._print_error()
            return None

        spike_array_shape = (schannel_info.m_NumberOfSpikes, schannel_info.m_SamplesPerSpike)

        return to_array(spike_timestamps), to_array(units), to_array(values).reshape(spike_array_shape)

    def pl2_get_digital_channel_info(self, zero_based_channel_index):
        """
        Retrieve information about a digital event channel

        Args:
            zero_based_channel_index - zero-based digital event channel index

        Returns:
            pl2_digital_channel_info - PL2DigitalChannelInfo class instance
        """

        self.pl2_dll.PL2_GetDigitalChannelInfo.argtypes = (
            ctypes.c_int,
            ctypes.c_int,
            ctypes.POINTER(PL2DigitalChannelInfo),
        )

        pl2_digital_channel_info = PL2DigitalChannelInfo()

        result = self.pl2_dll.PL2_GetDigitalChannelInfo(
            self._file_handle, ctypes.c_int(zero_based_channel_index), ctypes.byref(pl2_digital_channel_info)
        )

        if not result:
            self._print_error()
            return None

        return pl2_digital_channel_info

    def pl2_get_digital_channel_info_by_name(self, channel_name):
        """
        Retrieve information about a digital event channel

        Args:
            channel_name - digital event channel name

        Returns:
            pl2_digital_channel_info - PL2DigitalChannelInfo class instance
        """

        if hasattr(channel_name, "encode"):
            channel_name = channel_name.encode("ascii")

        self.pl2_dll.PL2_GetDigitalChannelInfoByName.argtypes = (
            ctypes.c_int,
            ctypes.POINTER(ctypes.c_char),
            ctypes.POINTER(PL2DigitalChannelInfo),
        )

        self.pl2_dll.PL2_GetDigitalChannelInfoByName.memsync = [
            {
                "p": [1],  # ctypes.POINTER argument
                "n": True,  # null-terminated string flag
            }
        ]

        pl2_digital_channel_info = PL2DigitalChannelInfo()

        result = self.pl2_dll.PL2_GetDigitalChannelInfoByName(
            self._file_handle, channel_name, ctypes.byref(pl2_digital_channel_info)
        )

        if not result:
            self._print_error()
            return None

        return pl2_digital_channel_info

    def pl2_get_digital_channel_info_by_source(self, source_id, one_based_channel_index_in_source):
        """
        Retrieve information about a digital event channel

        Args:
            source_id - numeric source ID
            one_based_channel_index_in_source - one-based channel index within the source

        Returns:
            pl2_digital_channel_info - PL2DigitalChannelInfo class instance
        """

        self.pl2_dll.PL2_GetDigitalChannelInfoBySource.argtypes = (
            ctypes.c_int,
            ctypes.c_int,
            ctypes.c_int,
            ctypes.POINTER(PL2DigitalChannelInfo),
        )

        pl2_digital_channel_info = PL2DigitalChannelInfo()

        result = self.pl2_dll.PL2_GetDigitalChannelInfoBySource(
            self._file_handle,
            ctypes.c_int(source_id),
            ctypes.c_int(one_based_channel_index_in_source),
            ctypes.byref(pl2_digital_channel_info),
        )

        if not result:
            self._print_error()
            return None

        return pl2_digital_channel_info

    def pl2_get_digital_channel_data(self, zero_based_channel_index):
        """
        Retrieve digital even channel data

        Args:
            zero_based_channel_index - zero-based digital event channel index

        Returns:
            event_timestamps - array the size of PL2DigitalChannelInfo.m_NumberOfEvents
            event_values - array the size of PL2DigitalChannelInfo.m_NumberOfEvents
        """

        self.pl2_dll.PL2_GetDigitalChannelData.argtypes = (
            ctypes.c_int,
            ctypes.c_int,
            ctypes.POINTER(ctypes.c_ulonglong),
            ctypes.POINTER(ctypes.c_longlong),
            ctypes.POINTER(ctypes.c_short),
        )

        self.pl2_dll.PL2_GetDigitalChannelData.memsync = [
            {"p": [3], "l": [2], "t": ctypes.c_longlong},
            {"p": [4], "l": [2], "t": ctypes.c_ushort},
        ]

        echannel_info = self.pl2_get_digital_channel_info(zero_based_channel_index)

        # These will be filled in by the dll method.
        num_events_returned = ctypes.c_ulonglong(echannel_info.m_NumberOfEvents)
        event_timestamps = (ctypes.c_longlong * echannel_info.m_NumberOfEvents)()
        event_values = (ctypes.c_ushort * echannel_info.m_NumberOfEvents)()

        result = self.pl2_dll.PL2_GetDigitalChannelData(
            self._file_handle,
            ctypes.c_int(zero_based_channel_index),
            num_events_returned,
            event_timestamps,
            event_values,
        )

        if not result:
            self._print_error()
            return None

        return to_array(event_timestamps), to_array(event_values)

    def pl2_get_digital_channel_data_by_name(self, channel_name):
        """
        Retrieve digital even channel data

        Args:
            channel_name - digital event channel name

        Returns:
            event_timestamps - array the size of PL2DigitalChannelInfo.m_NumberOfEvents
            event_values - array the size of PL2DigitalChannelInfo.m_NumberOfEvents
        """

        if hasattr(channel_name, "encode"):
            channel_name = channel_name.encode("ascii")

        self.pl2_dll.PL2_GetDigitalChannelDataByName.argtypes = (
            ctypes.c_int,  # file handle
            ctypes.POINTER(ctypes.c_char),  # channel name
            ctypes.POINTER(ctypes.c_ulonglong),  # equivalent to m_NumberOfEvents
            ctypes.POINTER(ctypes.c_longlong),  # array of timestamps of length m_NumberOfEvents
            ctypes.POINTER(ctypes.c_ushort),  # array of values m_NumberOfEvents
        )

        self.pl2_dll.PL2_GetDigitalChannelDataByName.memsync = [
            {
                "p": [1],  # ctypes.POINTER argument
                "n": True,  # null-terminated string flag
            },
            {"p": [3], "l": [2], "t": ctypes.c_longlong},
            {"p": [4], "l": [2], "t": ctypes.c_ushort},
        ]

        echannel_info = self.pl2_get_digital_channel_info_by_name(channel_name)

        # These will be filled in by the dll method.
        num_events_returned = ctypes.c_ulonglong(echannel_info.m_NumberOfEvents)
        event_timestamps = (ctypes.c_longlong * echannel_info.m_NumberOfEvents)()
        event_values = (ctypes.c_ushort * echannel_info.m_NumberOfEvents)()

        result = self.pl2_dll.PL2_GetDigitalChannelDataByName(
            self._file_handle, channel_name, num_events_returned, event_timestamps, event_values
        )

        if not result:
            self._print_error()
            return None

        return to_array(event_timestamps), to_array(event_values)

    def pl2_get_digital_channel_data_by_source(self, source_id, one_based_channel_index_in_source):
        """
        Retrieve digital even channel data

        Args:
            source_id - numeric source ID
            one_based_channel_index_in_source - one-based channel index within the source

        Returns:
            event_timestamps - array the size of PL2DigitalChannelInfo.m_NumberOfEvents
            event_values - array the size of PL2DigitalChannelInfo.m_NumberOfEvents
        """

        self.pl2_dll.PL2_GetDigitalChannelDataBySource.argtypes = (
            ctypes.c_int,
            ctypes.c_int,
            ctypes.c_int,
            ctypes.POINTER(ctypes.c_ulonglong),
            ctypes.POINTER(ctypes.c_longlong),
            ctypes.POINTER(ctypes.c_ushort),
        )

        self.pl2_dll.PL2_GetDigitalChannelDataBySource.memsync = [
            {"p": [4], "l": [3], "t": ctypes.c_longlong},
            {"p": [5], "l": [3], "t": ctypes.c_ushort},
        ]

        echannel_info = self.pl2_get_digital_channel_info_by_source(source_id, one_based_channel_index_in_source)

        # These will be filled in by the dll method.
        num_events_returned = ctypes.c_ulonglong(echannel_info.m_NumberOfEvents)
        event_timestamps = (ctypes.c_longlong * echannel_info.m_NumberOfEvents)()
        event_values = (ctypes.c_ushort * echannel_info.m_NumberOfEvents)()

        result = self.pl2_dll.PL2_GetDigitalChannelDataBySource(
            self._file_handle,
            ctypes.c_int(source_id),
            ctypes.c_int(one_based_channel_index_in_source),
            num_events_returned,
            event_timestamps,
            event_values,
        )

        if not result:
            self._print_error()
            return None

        return to_array(event_timestamps), to_array(event_values)

    def pl2_get_start_stop_channel_info(self, number_of_start_stop_events):
        """
        Retrieve information about start/stop channel

        Args:
            _file_handle - file handle
            number_of_start_stop_events - ctypes.c_ulonglong class instance

        Returns:
            1 - Success
            0 - Failure
            The class instances passed to the function are filled with values
        """

        self.pl2_dll.PL2_GetStartStopChannelInfo.argtypes = (ctypes.c_int, ctypes.POINTER(ctypes.c_ulonglong))

        result = self.pl2_dll.PL2_GetStartStopChannelInfo(self._file_handle, number_of_start_stop_events)

        return result

    def pl2_get_start_stop_channel_data(self, num_events_returned, event_timestamps, event_values):
        """
        Retrieve digital channel data

        Args:
            _file_handle - file handle
            num_events_returned - ctypes.c_ulonglong class instance
            event_timestamps - ctypes.c_longlong class instance
            event_values - point to ctypes.c_ushort class instance

        Returns:
            1 - Success
            0 - Failure
            The class instances passed to the function are filled with values
        """

        self.pl2_dll.PL2_GetStartStopChannelInfo.argtypes = (
            ctypes.c_int,
            ctypes.POINTER(ctypes.c_ulonglong),
            ctypes.POINTER(ctypes.c_longlong),
            ctypes.POINTER(ctypes.c_ushort),
        )

        self.pl2_dll.PL2_GetStartStopChannelInfo.memsync = [
            {"p": [2], "l": [1], "t": ctypes.c_longlong},
            {"p": [3], "l": [1], "t": ctypes.c_ushort},
        ]

        result = self.pl2_dll.PL2_GetStartStopChannelData(
            self._file_handle, num_events_returned, event_timestamps, event_values
        )

        return result

    def _print_error(self):
        error_message = self.pl2_get_last_error()
        print(f"pypl2lib error: {error_message}")

    # PL2 data block functions purposefully not implemented.
    def pl2_read_first_data_block(self):
        pass

    def pl2_read_next_data_block(self):
        pass

    def pl2_get_data_block_info(self):
        pass

    def pl2_get_data_block_timestamps(self):
        pass

    def pl2_get_spike_data_block_units(self):
        pass

    def pl2_get_spike_data_block_waveforms(self):
        pass

    def pl2_get_analog_data_block_timestamp(self):
        pass

    def pl2_get_analog_data_block_values(self):
        pass

    def pl2_get_digital_data_block_timestamps(self):
        pass

    def pl2_get_start_stop_data_block_timestamps(self):
        pass

    def pl2_get_start_stop_data_block_values(self):
        pass



"""
Tests of the neo.core.container.Container class
"""

import unittest

import quantities as pq

import numpy as np

import neo.core
from neo.core import filters

try:
    from IPython.lib.pretty import pretty
except ImportError as err:
    HAVE_IPYTHON = False
else:
    HAVE_IPYTHON = True

from neo.core.container import Container, unique_objs


class Test_unique_objs(unittest.TestCase):
    """
    TestCase for unique_objs
    """

    def test_some(self):
        a = 1
        b = np.array([3.14159265, 3.1415])
        c = [1, "1", 2.3, "5 8"]
        d = {1, "2", "spam"}

        objs = [a, b, b, b, c, b, a, d, b, b, a, d, d, d, c, d, b, d, c, a]
        targ = [a, b, c, d]
        res = unique_objs(objs)
        self.assertEqual(targ, res)


class TestContainerNeo(unittest.TestCase):
    """
    TestCase to make sure basic initialization and methods work
    """

    @classmethod
    def setUpClass(cls):
        seg = neo.core.Segment()
        st1 = neo.core.SpikeTrain([1, 2] * pq.ms, t_stop=10)
        st1.annotate(test=5)
        st2 = neo.core.SpikeTrain([3, 4] * pq.ms, t_stop=10)
        st2.annotate(test=6)
        st2.annotate(name="st_num_1")
        st2.annotate(filt=6)
        st3 = neo.core.SpikeTrain([5, 6] * pq.ms, t_stop=10)
        st3.annotate(list=[1, 2])
        st3.annotate(dict={"key": 5})
        seg.spiketrains.append(st1)
        seg.spiketrains.append(st2)
        seg.spiketrains.append(st3)

        cls.seg = seg
        cls.st1 = st1
        cls.st2 = st2

    def test_init(self):
        """test to make sure initialization works properly"""
        container = Container(name="a container", description="this is a test")
        self.assertEqual(container.name, "a container")
        self.assertEqual(container.description, "this is a test")
        self.assertEqual(container.file_origin, None)

    def test__children(self):
        container = Container()
        self.assertEqual(container._parent_objects, ())

        self.assertEqual(container._parent_containers, ())

        self.assertEqual(container._parent_objects, ())
        self.assertEqual(container._parent_containers, ())

        self.assertEqual(container._container_child_objects, ())
        self.assertEqual(container._data_child_objects, ())

        self.assertEqual(container._repr_pretty_containers, ())

        self.assertEqual(container._child_objects, ())

        self.assertEqual(container._container_child_containers, ())
        self.assertEqual(container._data_child_containers, ())
        self.assertEqual(container._child_containers, ())

        self.assertEqual(container._child_objects, ())
        self.assertEqual(container._child_containers, ())

        self.assertEqual(container._single_children, ())
        self.assertEqual(container.data_children, ())
        self.assertEqual(container.container_children, ())
        self.assertEqual(container.children, ())
        self.assertEqual(container.parents, ())

        self.assertEqual(container.data_children_recur, ())
        self.assertEqual(container.container_children_recur, ())
        self.assertEqual(container.children_recur, ())

        self.assertEqual(container.filter(test=1), [])
        self.assertEqual(container.filter(data=True, container=False, test=1), [])
        self.assertEqual(container.filter(data=False, container=False, test=1), [])
        self.assertEqual(container.filter(data=True, container=True, test=1), [])
        self.assertEqual(container.filter(data=False, container=True, test=1), [])

        self.assertEqual(container.size, {})

        container.check_relationships()

    def test_filter(self):
        container = Container()
        self.assertRaises(TypeError, container.filter, "foo")

    def test_filter_results(self):
        """
        Tests FilterConditions correct results
        """
        self.assertEqual(self.st1.annotations, self.seg.filter(test=filters.Equals(5))[0].annotations)
        self.assertEqual(self.st1.annotations, self.seg.filter(test=filters.LessThan(6))[0].annotations)
        self.assertEqual(self.st1.annotations, self.seg.filter(test=filters.GreaterThan(4))[0].annotations)
        self.assertEqual(self.st1.annotations, self.seg.filter(test=filters.IsNot(1))[0].annotations)
        self.assertEqual(self.st1.annotations, self.seg.filter(test=filters.IsIn([1, 2, 5]))[0].annotations)
        self.assertEqual(self.st1.annotations, self.seg.filter(test=filters.InRange(1, 5))[0].annotations)
        self.assertEqual(self.st1.annotations, self.seg.filter(test=filters.GreaterThanOrEquals(5))[0].annotations)
        self.assertEqual(self.st1.annotations, self.seg.filter(test=filters.LessThanOrEquals(5))[0].annotations)

    def test_filter_equal(self):
        """
        Tests FilterCondition object "Equals".
        """
        self.assertEqual(1, len(self.seg.filter(test=filters.Equals(5))))
        self.assertEqual(0, len(self.seg.filter(test=filters.Equals(1))))
        self.assertEqual(1, len(self.seg.filter({"list": filters.Equals([1, 2])})))
        self.assertEqual(1, len(self.seg.filter(dict=filters.Equals({"key": 5}))))

    def test_filter_is_not(self):
        """
        Tests FilterCondition object "IsNot".
        """
        self.assertEqual(2, len(self.seg.filter(test=filters.IsNot(1))))
        self.assertEqual(1, len(self.seg.filter(test=filters.IsNot(5))))
        self.assertEqual(0, len(self.seg.filter([{"test": filters.IsNot(5)}, {"test": filters.IsNot(6)}])))

    def test_filter_less_than(self):
        """
        Tests FilterCondition object "LessThan".
        """
        self.assertEqual(0, len(self.seg.filter(test=filters.LessThan(5))))
        self.assertEqual(1, len(self.seg.filter(test=filters.LessThan(6))))
        self.assertEqual(2, len(self.seg.filter(test=filters.LessThan(7))))

    def test_filter_less_than_equal(self):
        """
        Tests FilterCondition object "LessThanEquals".
        """
        self.assertEqual(0, len(self.seg.filter(test=filters.LessThanOrEquals(4))))
        self.assertEqual(1, len(self.seg.filter(test=filters.LessThanOrEquals(5))))
        self.assertEqual(2, len(self.seg.filter(test=filters.LessThanOrEquals(6))))

    def test_filter_greater_than(self):
        """
        Tests FilterCondition object "GreaterThan".
        """
        self.assertEqual(0, len(self.seg.filter(test=filters.GreaterThan(6))))
        self.assertEqual(1, len(self.seg.filter(test=filters.GreaterThan(5))))
        self.assertEqual(2, len(self.seg.filter(test=filters.GreaterThan(4))))

    def test_filter_greater_than_equal(self):
        """
        Tests FilterCondition object "GreaterThanEquals".
        """
        self.assertEqual(0, len(self.seg.filter(test=filters.GreaterThanOrEquals(7))))
        self.assertEqual(1, len(self.seg.filter(test=filters.GreaterThanOrEquals(6))))
        self.assertEqual(2, len(self.seg.filter(test=filters.GreaterThanOrEquals(5))))

    def test_filter_is_in(self):
        """
        Tests FilterCondition object "IsIn".
        """
        # list
        self.assertEqual(0, len(self.seg.filter(test=filters.IsIn([4, 7, 10]))))
        self.assertEqual(1, len(self.seg.filter(test=filters.IsIn([5, 7, 10]))))
        self.assertEqual(2, len(self.seg.filter(test=filters.IsIn([5, 6, 10]))))
        # tuple
        self.assertEqual(0, len(self.seg.filter(test=filters.IsIn((4, 7, 10)))))
        self.assertEqual(1, len(self.seg.filter(test=filters.IsIn((5, 7, 10)))))
        self.assertEqual(2, len(self.seg.filter(test=filters.IsIn((5, 6, 10)))))
        # set
        self.assertEqual(0, len(self.seg.filter(test=filters.IsIn({4, 7, 10}))))
        self.assertEqual(1, len(self.seg.filter(test=filters.IsIn({5, 7, 10}))))
        self.assertEqual(2, len(self.seg.filter(test=filters.IsIn({5, 6, 10}))))

    def test_filter_in_range(self):
        """
        Tests FilterCondition object "InRange".
        """
        with self.assertRaises(ValueError):
            self.seg.filter(test=filters.InRange("wrong", 6, False, False))
        self.assertEqual(2, len(self.seg.filter(test=filters.InRange(5, 6, False, False))))
        self.assertEqual(1, len(self.seg.filter(test=filters.InRange(5, 6, True, False))))
        self.assertEqual(1, len(self.seg.filter(test=filters.InRange(5, 6, False, True))))
        self.assertEqual(0, len(self.seg.filter(test=filters.InRange(5, 6, True, True))))

    def test_filter_filter_consistency(self):
        """
        Tests old functionality with new filter method.
        """
        self.assertEqual(2, len(self.seg.filter({"test": filters.Equals(5), "filt": filters.Equals(6)})))
        self.assertEqual(0, len(self.seg.filter([{"test": filters.Equals(5)}, {"filt": filters.Equals(6)}])))
        self.assertEqual(1, len(self.seg.filter(name="st_num_1")))


class Test_Container_merge(unittest.TestCase):
    """
    TestCase to make sure merge method works
    """

    def setUp(self):
        self.name1 = "a container 1"
        self.name2 = "a container 2"
        self.description1 = "this is a test 1"
        self.description2 = "this is a test 2"
        self.cont1 = Container(name=self.name1, description=self.description1)
        self.cont2 = Container(name=self.name2, description=self.description2)

    def test_merge__dict(self):
        self.cont1.annotations = {"val1": 1, "val2": 2.2, "val3": "test1"}
        self.cont2.annotations = {"val2": 2.2, "val3": "test2", "val4": [4, 4.4], "val5": True}

        ann1 = self.cont1.annotations
        ann1c = self.cont1.annotations.copy()
        ann2c = self.cont2.annotations.copy()

        targ = {"val1": 1, "val2": 2.2, "val3": "test1;test2", "val4": [4, 4.4], "val5": True}

        self.cont1.merge(self.cont2)

        self.assertEqual(ann1, self.cont1.annotations)
        self.assertNotEqual(ann1c, self.cont1.annotations)
        self.assertEqual(ann2c, self.cont2.annotations)
        self.assertEqual(targ, self.cont1.annotations)

        self.assertEqual(self.name1, self.cont1.name)
        self.assertEqual(self.name2, self.cont2.name)
        self.assertEqual(self.description1, self.cont1.description)
        self.assertEqual(self.description2, self.cont2.description)

    def test_merge__different_type_AssertionError(self):
        self.cont1.annotations = {"val1": 1, "val2": 2.2, "val3": "tester"}
        self.cont2.annotations = {"val3": False, "val4": [4, 4.4], "val5": True}
        self.cont1.merge(self.cont2)
        self.assertEqual(
            self.cont1.annotations, {"val1": 1, "val2": 2.2, "val3": "MERGE CONFLICT", "val4": [4, 4.4], "val5": True}
        )

    def test_merge__unmergable_unequal_AssertionError(self):
        self.cont1.annotations = {"val1": 1, "val2": 2.2, "val3": True}
        self.cont2.annotations = {"val3": False, "val4": [4, 4.4], "val5": True}
        self.cont1.merge(self.cont2)
        self.assertEqual(
            self.cont1.annotations, {"val1": 1, "val2": 2.2, "val3": "MERGE CONFLICT", "val4": [4, 4.4], "val5": True}
        )


@unittest.skipUnless(HAVE_IPYTHON, "requires IPython")
class Test_pprint(unittest.TestCase):
    def test__pretty(self):
        name = "an object"
        description = "this is a test"
        obj = Container(name=name, description=description)
        res = pretty(obj)
        targ = f"Container with  name: '{name}' description: '{description}'"
        self.assertEqual(res, targ)


if __name__ == "__main__":
    unittest.main()



"""
Tests of neo.io.brainvisionio
"""

import unittest

from neo.io import BrainVisionIO

from neo.test.iotest.common_io_test import BaseTestIO


class TestBrainVisionIO(
    BaseTestIO,
    unittest.TestCase,
):
    ioclass = BrainVisionIO
    entities_to_download = ["brainvision"]
    entities_to_test = [
        "brainvision/File_brainvision_1.vhdr",
        "brainvision/File_brainvision_2.vhdr",
        "brainvision/File_brainvision_3_float32.vhdr",
        "brainvision/File_brainvision_3_int16.vhdr",
        "brainvision/File_brainvision_3_int32.vhdr",
    ]


if __name__ == "__main__":
    unittest.main()



"""
Tests of neo.io.axonio
"""

import unittest

from neo.io import AxonIO
from neo.test.iotest.common_io_test import BaseTestIO

from neo.test.rawiotest.test_axonrawio import TestAxonRawIO


class TestAxonIO(BaseTestIO, unittest.TestCase):
    entities_to_download = TestAxonRawIO.entities_to_download
    entities_to_test = TestAxonRawIO.entities_to_test
    ioclass = AxonIO

    def test_annotations(self):
        reader = AxonIO(filename=self.get_local_path("axon/File_axon_2.abf"))
        bl = reader.read_block()
        ev = bl.segments[0].events[0]
        assert "comments" in ev.annotations

    def test_read_protocol(self):
        for f in self.entities_to_test:
            filename = self.get_local_path(f)
            reader = AxonIO(filename=filename)
            bl = reader.read_block(lazy=True)
            if bl.annotations["abf_version"] >= 2.0:
                reader.read_protocol()


if __name__ == "__main__":
    unittest.main()



"""
Tests of the neo.utils.misc module
"""

import unittest
import warnings

import numpy as np
import quantities as pq
from neo.rawio.examplerawio import ExampleRawIO
from neo.io.proxyobjects import AnalogSignalProxy, SpikeTrainProxy, EventProxy, EpochProxy

from neo.core.dataobject import ArrayDict
from neo.core import Block, Segment, AnalogSignal, IrregularlySampledSignal, Epoch, Event, SpikeTrain

from neo.test.tools import (
    assert_arrays_almost_equal,
    assert_arrays_equal,
    assert_neo_object_is_compliant,
    assert_same_attributes,
    assert_same_annotations,
)

from neo.utils import get_events, get_epochs, add_epoch, match_events, cut_block_by_epochs


class BaseProxyTest(unittest.TestCase):
    def setUp(self):
        self.reader = ExampleRawIO(filename="my_filename.fake")
        self.reader.parse_header()


class TestUtilsWithoutProxyObjects(unittest.TestCase):
    def test__get_events(self):
        starts_1 = Event(times=[0.5, 10.0, 25.2] * pq.s, labels=["label1", "label2", "label3"], name="pick_me")
        starts_1.annotate(event_type="trial start")
        starts_1.array_annotate(trial_id=[1, 2, 3])

        stops_1 = Event(times=[5.5, 14.9, 30.1] * pq.s)
        stops_1.annotate(event_type="trial stop")
        stops_1.array_annotate(trial_id=[1, 2, 3])

        starts_2 = Event(times=[33.2, 41.7, 52.4] * pq.s)
        starts_2.annotate(event_type="trial start")
        starts_2.array_annotate(trial_id=[4, 5, 6])

        stops_2 = Event(times=[37.6, 46.1, 57.0] * pq.s)
        stops_2.annotate(event_type="trial stop")
        stops_2.array_annotate(trial_id=[4, 5, 6])

        seg = Segment()
        seg2 = Segment()
        seg.events = [starts_1, stops_1]
        seg2.events = [starts_2, stops_2]

        block = Block()
        block.segments = [seg, seg2]

        # test getting one whole event via annotation or attribute
        extracted_starts1 = get_events(seg, event_type="trial start")
        extracted_starts1b = get_events(block, name="pick_me")

        self.assertEqual(len(extracted_starts1), 1)
        self.assertEqual(len(extracted_starts1b), 1)

        extracted_starts1 = extracted_starts1[0]
        extracted_starts1b = extracted_starts1b[0]

        assert_same_attributes(extracted_starts1, starts_1)
        assert_same_attributes(extracted_starts1b, starts_1)

        # test getting an empty list by searching for a non-existent property
        empty1 = get_events(seg, foo="bar")

        self.assertEqual(len(empty1), 0)

        # test getting an empty list by searching for a non-existent property value
        empty2 = get_events(seg, event_type="undefined")

        self.assertEqual(len(empty2), 0)

        # test getting only one event time of one event
        trial_2 = get_events(block, trial_id=2, event_type="trial start")

        self.assertEqual(len(trial_2), 1)

        trial_2 = trial_2[0]

        self.assertEqual(starts_1.name, trial_2.name)
        self.assertEqual(starts_1.description, trial_2.description)
        self.assertEqual(starts_1.file_origin, trial_2.file_origin)
        self.assertEqual(starts_1.annotations["event_type"], trial_2.annotations["event_type"])
        assert_arrays_equal(trial_2.array_annotations["trial_id"], np.array([2]))
        self.assertIsInstance(trial_2.array_annotations, ArrayDict)

        # test getting only one event time of more than one event
        trial_2b = get_events(block, trial_id=2)

        self.assertEqual(len(trial_2b), 2)

        start_idx = np.where(np.array([ev.annotations["event_type"] for ev in trial_2b]) == "trial start")[0][0]

        trial_2b_start = trial_2b[start_idx]
        trial_2b_stop = trial_2b[start_idx - 1]

        assert_same_attributes(trial_2b_start, trial_2)

        self.assertEqual(stops_1.name, trial_2b_stop.name)
        self.assertEqual(stops_1.description, trial_2b_stop.description)
        self.assertEqual(stops_1.file_origin, trial_2b_stop.file_origin)
        self.assertEqual(stops_1.annotations["event_type"], trial_2b_stop.annotations["event_type"])
        assert_arrays_equal(trial_2b_stop.array_annotations["trial_id"], np.array([2]))
        self.assertIsInstance(trial_2b_stop.array_annotations, ArrayDict)

        # test getting more than one event time of one event
        trials_1_2 = get_events(block, trial_id=[1, 2], event_type="trial start")

        self.assertEqual(len(trials_1_2), 1)

        trials_1_2 = trials_1_2[0]

        self.assertEqual(starts_1.name, trials_1_2.name)
        self.assertEqual(starts_1.description, trials_1_2.description)
        self.assertEqual(starts_1.file_origin, trials_1_2.file_origin)
        self.assertEqual(starts_1.annotations["event_type"], trials_1_2.annotations["event_type"])
        assert_arrays_equal(trials_1_2.array_annotations["trial_id"], np.array([1, 2]))
        self.assertIsInstance(trials_1_2.array_annotations, ArrayDict)

        # test selecting event times by label
        trials_1_2 = get_events(block, labels=["label1", "label2"])

        self.assertEqual(len(trials_1_2), 1)

        trials_1_2 = trials_1_2[0]

        self.assertEqual(starts_1.name, trials_1_2.name)
        self.assertEqual(starts_1.description, trials_1_2.description)
        self.assertEqual(starts_1.file_origin, trials_1_2.file_origin)
        self.assertEqual(starts_1.annotations["event_type"], trials_1_2.annotations["event_type"])
        assert_arrays_equal(trials_1_2.array_annotations["trial_id"], np.array([1, 2]))
        self.assertIsInstance(trials_1_2.array_annotations, ArrayDict)

        # test getting more than one event time of more than one event
        trials_1_2b = get_events(block, trial_id=[1, 2])

        self.assertEqual(len(trials_1_2b), 2)

        start_idx = np.where(np.array([ev.annotations["event_type"] for ev in trials_1_2b]) == "trial start")[0][0]

        trials_1_2b_start = trials_1_2b[start_idx]
        trials_1_2b_stop = trials_1_2b[start_idx - 1]

        assert_same_attributes(trials_1_2b_start, trials_1_2)

        self.assertEqual(stops_1.name, trials_1_2b_stop.name)
        self.assertEqual(stops_1.description, trials_1_2b_stop.description)
        self.assertEqual(stops_1.file_origin, trials_1_2b_stop.file_origin)
        self.assertEqual(stops_1.annotations["event_type"], trials_1_2b_stop.annotations["event_type"])
        assert_arrays_equal(trials_1_2b_stop.array_annotations["trial_id"], np.array([1, 2]))
        self.assertIsInstance(trials_1_2b_stop.array_annotations, ArrayDict)

    def test__get_epochs(self):
        a_1 = Epoch([0.5, 10.0, 25.2] * pq.s, durations=[5.1, 4.8, 5.0] * pq.s)
        a_1.annotate(epoch_type="a", pick="me")
        a_1.array_annotate(trial_id=[1, 2, 3])

        b_1 = Epoch([5.5, 14.9, 30.1] * pq.s, durations=[4.7, 4.9, 5.2] * pq.s)
        b_1.annotate(epoch_type="b")
        b_1.array_annotate(trial_id=[1, 2, 3])

        a_2 = Epoch([33.2, 41.7, 52.4] * pq.s, durations=[5.3, 5.0, 5.1] * pq.s)
        a_2.annotate(epoch_type="a")
        a_2.array_annotate(trial_id=[4, 5, 6])

        b_2 = Epoch([37.6, 46.1, 57.0] * pq.s, durations=[4.9, 5.2, 5.1] * pq.s)
        b_2.annotate(epoch_type="b")
        b_2.array_annotate(trial_id=[4, 5, 6])

        seg = Segment()
        seg2 = Segment()
        seg.epochs = [a_1, b_1]
        seg2.epochs = [a_2, b_2]

        block = Block()
        block.segments = [seg, seg2]

        # test getting one whole event via annotation
        extracted_a_1 = get_epochs(seg, epoch_type="a")
        extracted_a_1b = get_epochs(block, pick="me")

        self.assertEqual(len(extracted_a_1), 1)
        self.assertEqual(len(extracted_a_1b), 1)

        extracted_a_1 = extracted_a_1[0]
        extracted_a_1b = extracted_a_1b[0]

        assert_same_attributes(extracted_a_1, a_1)
        assert_same_attributes(extracted_a_1b, a_1)

        # test getting an empty list by searching for a non-existent property
        empty1 = get_epochs(seg, foo="bar")

        self.assertEqual(len(empty1), 0)

        # test getting an empty list by searching for a non-existent property value
        empty2 = get_epochs(seg, epoch_type="undefined")

        self.assertEqual(len(empty2), 0)

        # test getting only one event time of one event
        trial_2 = get_epochs(block, trial_id=2, epoch_type="a")

        self.assertEqual(len(trial_2), 1)

        trial_2 = trial_2[0]

        self.assertEqual(a_1.name, trial_2.name)
        self.assertEqual(a_1.description, trial_2.description)
        self.assertEqual(a_1.file_origin, trial_2.file_origin)
        self.assertEqual(a_1.annotations["epoch_type"], trial_2.annotations["epoch_type"])
        assert_arrays_equal(trial_2.array_annotations["trial_id"], np.array([2]))
        self.assertIsInstance(trial_2.array_annotations, ArrayDict)

        # test getting only one event time of more than one event
        trial_2b = get_epochs(block, trial_id=2)

        self.assertEqual(len(trial_2b), 2)

        a_idx = np.where(np.array([ev.annotations["epoch_type"] for ev in trial_2b]) == "a")[0][0]

        trial_2b_a = trial_2b[a_idx]
        trial_2b_b = trial_2b[a_idx - 1]

        assert_same_attributes(trial_2b_a, trial_2)

        self.assertEqual(b_1.name, trial_2b_b.name)
        self.assertEqual(b_1.description, trial_2b_b.description)
        self.assertEqual(b_1.file_origin, trial_2b_b.file_origin)
        self.assertEqual(b_1.annotations["epoch_type"], trial_2b_b.annotations["epoch_type"])
        assert_arrays_equal(trial_2b_b.array_annotations["trial_id"], np.array([2]))
        self.assertIsInstance(trial_2b_b.array_annotations, ArrayDict)

        # test getting more than one event time of one event
        trials_1_2 = get_epochs(block, trial_id=[1, 2], epoch_type="a")

        self.assertEqual(len(trials_1_2), 1)

        trials_1_2 = trials_1_2[0]

        self.assertEqual(a_1.name, trials_1_2.name)
        self.assertEqual(a_1.description, trials_1_2.description)
        self.assertEqual(a_1.file_origin, trials_1_2.file_origin)
        self.assertEqual(a_1.annotations["epoch_type"], trials_1_2.annotations["epoch_type"])
        assert_arrays_equal(trials_1_2.array_annotations["trial_id"], np.array([1, 2]))
        self.assertIsInstance(trials_1_2.array_annotations, ArrayDict)

        # test getting more than one event time of more than one event
        trials_1_2b = get_epochs(block, trial_id=[1, 2])

        self.assertEqual(len(trials_1_2b), 2)

        a_idx = np.where(np.array([ev.annotations["epoch_type"] for ev in trials_1_2b]) == "a")[0][0]

        trials_1_2b_a = trials_1_2b[a_idx]
        trials_1_2b_b = trials_1_2b[a_idx - 1]

        assert_same_attributes(trials_1_2b_a, trials_1_2)

        self.assertEqual(b_1.name, trials_1_2b_b.name)
        self.assertEqual(b_1.description, trials_1_2b_b.description)
        self.assertEqual(b_1.file_origin, trials_1_2b_b.file_origin)
        self.assertEqual(b_1.annotations["epoch_type"], trials_1_2b_b.annotations["epoch_type"])
        assert_arrays_equal(trials_1_2b_b.array_annotations["trial_id"], np.array([1, 2]))
        self.assertIsInstance(trials_1_2b_b.array_annotations, ArrayDict)

    def test__add_epoch(self):
        starts = Event(times=[0.5, 10.0, 25.2] * pq.s)
        starts.annotate(event_type="trial start", nix_name="neo.event.0")
        starts.array_annotate(trial_id=[1, 2, 3])

        stops = Event(times=[5.5, 14.9, 30.1] * pq.s)
        stops.annotate(event_type="trial stop", nix_name="neo.event.1")
        stops.array_annotate(trial_id=[1, 2, 3])

        seg = Segment()
        seg.events = [starts, stops]

        # test cutting with one event only
        ep_starts = add_epoch(seg, starts, pre=-300 * pq.ms, post=250 * pq.ms)

        assert_neo_object_is_compliant(ep_starts)
        self.assertDictEqual(ep_starts.annotations, {"event_type": "trial start"})
        assert_arrays_almost_equal(ep_starts.times, starts.times - 300 * pq.ms, 1e-12)
        assert_arrays_almost_equal(
            ep_starts.durations, (550 * pq.ms).rescale(ep_starts.durations.units) * np.ones(len(starts)), 1e-12
        )

        # test cutting with two events
        ep_trials = add_epoch(seg, starts, stops)

        assert_neo_object_is_compliant(ep_trials)
        self.assertDictEqual(ep_trials.annotations, {"event_type": "trial start"})
        assert_arrays_almost_equal(ep_trials.times, starts.times, 1e-12)
        assert_arrays_almost_equal(ep_trials.durations, stops - starts, 1e-12)

    def test__match_events(self):
        starts = Event(times=[0.5, 10.0, 25.2] * pq.s)
        starts.annotate(event_type="trial start")
        starts.array_annotate(trial_id=[1, 2, 3])

        stops = Event(times=[5.5, 14.9, 30.1] * pq.s)
        stops.annotate(event_type="trial stop")
        stops.array_annotate(trial_id=[1, 2, 3])

        stops2 = Event(times=[0.1, 5.5, 5.6, 14.9, 25.2, 30.1] * pq.s)
        stops2.annotate(event_type="trial stop")
        stops2.array_annotate(trial_id=[1, 1, 2, 2, 3, 3])

        # test for matching input events, should just return identical copies
        matched_starts, matched_stops = match_events(starts, stops)

        assert_same_attributes(matched_starts, starts)
        assert_same_attributes(matched_stops, stops)

        # test for non-matching input events, should find shortest positive non-zero durations
        matched_starts2, matched_stops2 = match_events(starts, stops2)

        assert_same_attributes(matched_starts2, starts)
        assert_same_attributes(matched_stops2, stops)

    def test__cut_block_by_epochs(self):
        epoch = Epoch([0.5, 10.0, 25.2] * pq.s, durations=[5.1, 4.8, 5.0] * pq.s, t_start=0.1 * pq.s)
        epoch.annotate(epoch_type="a", pick="me", nix_name="neo.epoch.0")
        epoch.array_annotate(trial_id=[1, 2, 3])

        epoch2 = Epoch([0.6, 9.5, 16.8, 34.1] * pq.s, durations=[4.5, 4.8, 5.0, 5.0] * pq.s, t_start=0.1 * pq.s)
        epoch2.annotate(epoch_type="b", nix_name="neo.epoch.1")
        epoch2.array_annotate(trial_id=[1, 2, 3, 4])

        event = Event(times=[0.5, 10.0, 25.2] * pq.s, t_start=0.1 * pq.s)
        event.annotate(event_type="trial start", nix_name="neo.event.0")
        event.array_annotate(trial_id=[1, 2, 3])

        anasig = AnalogSignal(np.arange(50.0) * pq.mV, t_start=0.1 * pq.s, sampling_rate=1.0 * pq.Hz)
        irrsig = IrregularlySampledSignal(signal=np.arange(50.0) * pq.mV, times=anasig.times, t_start=0.1 * pq.s)
        st = SpikeTrain(
            np.arange(0.5, 50, 7) * pq.s,
            t_start=0.1 * pq.s,
            t_stop=50.0 * pq.s,
            waveforms=np.array(
                [
                    [[0.0, 1.0], [0.1, 1.1]],
                    [[2.0, 3.0], [2.1, 3.1]],
                    [[4.0, 5.0], [4.1, 5.1]],
                    [[6.0, 7.0], [6.1, 7.1]],
                    [[8.0, 9.0], [8.1, 9.1]],
                    [[12.0, 13.0], [12.1, 13.1]],
                    [[14.0, 15.0], [14.1, 15.1]],
                    [[16.0, 17.0], [16.1, 17.1]],
                ]
            )
            * pq.mV,
            array_annotations={"spikenum": np.arange(1, 9)},
        )

        # test without resetting the time
        seg = Segment(nix_name="neo.segment.0")
        seg2 = Segment(name="NoCut", nix_name="neo.segment.1")
        seg.epochs = [epoch, epoch2]
        seg.events = [event]
        seg.analogsignals = [anasig]
        seg.irregularlysampledsignals = [irrsig]
        seg.spiketrains = [st]

        original_block = Block()
        original_block.segments = [seg, seg2]
        original_block.check_relationships()

        with warnings.catch_warnings(record=True) as w:
            # This should raise a warning as one segment does not contain epochs
            block = cut_block_by_epochs(original_block, properties={"pick": "me"})
            self.assertEqual(len(w), 1)

        assert_neo_object_is_compliant(block)
        self.assertEqual(len(block.segments), 3)

        for epoch_idx in range(len(epoch)):
            self.assertEqual(len(block.segments[epoch_idx].events), 1)
            self.assertEqual(len(block.segments[epoch_idx].spiketrains), 1)
            self.assertEqual(len(block.segments[epoch_idx].analogsignals), 1)
            self.assertEqual(len(block.segments[epoch_idx].irregularlysampledsignals), 1)

            annos = block.segments[epoch_idx].annotations
            # new segment objects have different identity
            self.assertNotIn("nix_name", annos)

            if epoch_idx != 0:
                self.assertEqual(len(block.segments[epoch_idx].epochs), 1)
            else:
                self.assertEqual(len(block.segments[epoch_idx].epochs), 2)

            assert_same_attributes(
                block.segments[epoch_idx].spiketrains[0],
                st.time_slice(
                    t_start=epoch.times[epoch_idx], t_stop=epoch.times[epoch_idx] + epoch.durations[epoch_idx]
                ),
            )
            assert_same_attributes(
                block.segments[epoch_idx].analogsignals[0],
                anasig.time_slice(
                    t_start=epoch.times[epoch_idx], t_stop=epoch.times[epoch_idx] + epoch.durations[epoch_idx]
                ),
            )
            assert_same_attributes(
                block.segments[epoch_idx].irregularlysampledsignals[0],
                irrsig.time_slice(
                    t_start=epoch.times[epoch_idx], t_stop=epoch.times[epoch_idx] + epoch.durations[epoch_idx]
                ),
            )
            assert_same_attributes(
                block.segments[epoch_idx].events[0],
                event.time_slice(
                    t_start=epoch.times[epoch_idx], t_stop=epoch.times[epoch_idx] + epoch.durations[epoch_idx]
                ),
            )
        assert_same_attributes(
            block.segments[0].epochs[0],
            epoch.time_slice(t_start=epoch.times[0], t_stop=epoch.times[0] + epoch.durations[0]),
        )
        assert_same_attributes(
            block.segments[0].epochs[1],
            epoch2.time_slice(t_start=epoch.times[0], t_stop=epoch.times[0] + epoch.durations[0]),
        )

        # test with resetting the time
        seg = Segment(nix_name="neo.segment.0")
        seg2 = Segment(name="NoCut", nix_name="neo.segment.1")
        seg.epochs = [epoch, epoch2]
        seg.events = [event]
        seg.analogsignals = [anasig]
        seg.irregularlysampledsignals = [irrsig]
        seg.spiketrains = [st]

        original_block = Block()
        original_block.segments = [seg, seg2]
        original_block.check_relationships()

        with warnings.catch_warnings(record=True) as w:
            # This should raise a warning as one segment does not contain epochs
            block = cut_block_by_epochs(original_block, properties={"pick": "me"}, reset_time=True)
            self.assertEqual(len(w), 1)

        assert_neo_object_is_compliant(block)
        self.assertEqual(len(block.segments), 3)

        for epoch_idx in range(len(epoch)):
            self.assertEqual(len(block.segments[epoch_idx].events), 1)
            self.assertEqual(len(block.segments[epoch_idx].spiketrains), 1)
            self.assertEqual(len(block.segments[epoch_idx].analogsignals), 1)
            self.assertEqual(len(block.segments[epoch_idx].irregularlysampledsignals), 1)

            annos = block.segments[epoch_idx].annotations
            self.assertNotIn("nix_name", annos)

            if epoch_idx != 0:
                self.assertEqual(len(block.segments[epoch_idx].epochs), 1)
            else:
                self.assertEqual(len(block.segments[epoch_idx].epochs), 2)

            assert_same_attributes(
                block.segments[epoch_idx].spiketrains[0],
                st.time_shift(-epoch.times[epoch_idx]).time_slice(t_start=0 * pq.s, t_stop=epoch.durations[epoch_idx]),
            )

            anasig_target = anasig.time_shift(-epoch.times[epoch_idx])
            anasig_target = anasig_target.time_slice(t_start=0 * pq.s, t_stop=epoch.durations[epoch_idx])
            assert_same_attributes(block.segments[epoch_idx].analogsignals[0], anasig_target)
            irrsig_target = irrsig.time_shift(-epoch.times[epoch_idx])
            irrsig_target = irrsig_target.time_slice(t_start=0 * pq.s, t_stop=epoch.durations[epoch_idx])
            assert_same_attributes(block.segments[epoch_idx].irregularlysampledsignals[0], irrsig_target)
            assert_same_attributes(
                block.segments[epoch_idx].events[0],
                event.time_shift(-epoch.times[epoch_idx]).time_slice(
                    t_start=0 * pq.s, t_stop=epoch.durations[epoch_idx]
                ),
            )

        assert_same_attributes(
            block.segments[0].epochs[0],
            epoch.time_shift(-epoch.times[0]).time_slice(t_start=0 * pq.s, t_stop=epoch.durations[0]),
        )
        assert_same_attributes(
            block.segments[0].epochs[1],
            epoch2.time_shift(-epoch.times[0]).time_slice(t_start=0 * pq.s, t_stop=epoch.durations[0]),
        )


class TestUtilsWithProxyObjects(BaseProxyTest):
    def test__get_events(self):
        starts_1 = Event(times=[0.5, 10.0, 25.2] * pq.s)
        starts_1.annotate(event_type="trial start", pick="me")
        starts_1.array_annotate(trial_id=[1, 2, 3])

        stops_1 = Event(times=[5.5, 14.9, 30.1] * pq.s)
        stops_1.annotate(event_type="trial stop")
        stops_1.array_annotate(trial_id=[1, 2, 3])

        proxy_event = EventProxy(rawio=self.reader, event_channel_index=0, block_index=0, seg_index=0)

        proxy_event.annotate(event_type="trial start")

        seg = Segment()
        seg.events = [starts_1, stops_1, proxy_event]

        # test getting multiple events including a proxy
        extracted_starts = get_events(seg, event_type="trial start")

        self.assertEqual(len(extracted_starts), 2)

        # make sure the event is loaded and a neo.Event object is returned
        self.assertTrue(isinstance(extracted_starts[0], Event))
        self.assertTrue(isinstance(extracted_starts[1], Event))

    def test__get_epochs(self):
        a = Epoch([0.5, 10.0, 25.2] * pq.s, durations=[5.1, 4.8, 5.0] * pq.s)
        a.annotate(epoch_type="a", pick="me")
        a.array_annotate(trial_id=[1, 2, 3])

        b = Epoch([5.5, 14.9, 30.1] * pq.s, durations=[4.7, 4.9, 5.2] * pq.s)
        b.annotate(epoch_type="b")
        b.array_annotate(trial_id=[1, 2, 3])

        proxy_epoch = EpochProxy(rawio=self.reader, event_channel_index=1, block_index=0, seg_index=0)

        proxy_epoch.annotate(epoch_type="a")

        seg = Segment()
        seg.epochs = [a, b, proxy_epoch]

        # test getting multiple epochs including a proxy
        extracted_epochs = get_epochs(seg, epoch_type="a")

        self.assertEqual(len(extracted_epochs), 2)

        # make sure the epoch is loaded and a neo.Epoch object is returned
        self.assertTrue(isinstance(extracted_epochs[0], Epoch))
        self.assertTrue(isinstance(extracted_epochs[1], Epoch))

    def test__add_epoch(self):
        proxy_event = EventProxy(rawio=self.reader, event_channel_index=0, block_index=0, seg_index=0)

        loaded_event = proxy_event.load()

        regular_event = Event(times=loaded_event.times - 1 * loaded_event.units)

        loaded_event.annotate(nix_name="neo.event.0")
        regular_event.annotate(nix_name="neo.event.1")

        seg = Segment()
        seg.events = [regular_event, proxy_event]

        # test cutting with two events one of which is a proxy
        epoch = add_epoch(seg, regular_event, proxy_event)

        assert_neo_object_is_compliant(epoch)
        exp_annos = {k: v for k, v in regular_event.annotations.items() if k != "nix_name"}
        self.assertDictEqual(epoch.annotations, exp_annos)
        assert_arrays_almost_equal(epoch.times, regular_event.times, 1e-12)
        assert_arrays_almost_equal(epoch.durations, np.ones(regular_event.shape) * loaded_event.units, 1e-12)

    def test__match_events(self):
        proxy_event = EventProxy(rawio=self.reader, event_channel_index=0, block_index=0, seg_index=0)

        loaded_event = proxy_event.load()

        regular_event = Event(
            times=loaded_event.times - 1 * loaded_event.units,
            labels=np.array(["trigger_a", "trigger_b"] * 3, dtype="U12"),
        )

        seg = Segment()
        seg.events = [regular_event, proxy_event]

        # test matching two events one of which is a proxy
        matched_regular, matched_proxy = match_events(regular_event, proxy_event)

        assert_same_attributes(matched_regular, regular_event)
        assert_same_attributes(matched_proxy, loaded_event)

    def test__cut_block_by_epochs(self):
        seg = Segment()

        proxy_anasig = AnalogSignalProxy(
            rawio=self.reader, stream_index=0, inner_stream_channels=None, block_index=0, seg_index=0
        )
        seg.analogsignals.append(proxy_anasig)

        proxy_st = SpikeTrainProxy(rawio=self.reader, spike_channel_index=0, block_index=0, seg_index=0)
        seg.spiketrains.append(proxy_st)

        proxy_event = EventProxy(rawio=self.reader, event_channel_index=0, block_index=0, seg_index=0)
        seg.events.append(proxy_event)

        proxy_epoch = EpochProxy(rawio=self.reader, event_channel_index=1, block_index=0, seg_index=0)
        proxy_epoch.annotate(pick="me")
        seg.epochs.append(proxy_epoch)

        loaded_epoch = proxy_epoch.load()
        loaded_event = proxy_event.load()
        loaded_st = proxy_st.load()
        loaded_anasig = proxy_anasig.load()

        original_block = Block()
        original_block.segments = [seg]
        original_block.check_relationships()

        block = cut_block_by_epochs(original_block, properties={"pick": "me"})

        assert_neo_object_is_compliant(block)
        self.assertEqual(len(block.segments), proxy_epoch.shape[0])

        for epoch_idx in range(len(loaded_epoch)):
            sliced_event = loaded_event.time_slice(
                t_start=loaded_epoch.times[epoch_idx],
                t_stop=loaded_epoch.times[epoch_idx] + loaded_epoch.durations[epoch_idx],
            )
            has_event = len(sliced_event) > 0

            sliced_anasig = loaded_anasig.time_slice(
                t_start=loaded_epoch.times[epoch_idx],
                t_stop=loaded_epoch.times[epoch_idx] + loaded_epoch.durations[epoch_idx],
            )

            sliced_st = loaded_st.time_slice(
                t_start=loaded_epoch.times[epoch_idx],
                t_stop=loaded_epoch.times[epoch_idx] + loaded_epoch.durations[epoch_idx],
            )

            self.assertEqual(len(block.segments[epoch_idx].events), int(has_event))
            self.assertEqual(len(block.segments[epoch_idx].spiketrains), 1)
            self.assertEqual(len(block.segments[epoch_idx].analogsignals), 1)

            self.assertTrue(isinstance(block.segments[epoch_idx].spiketrains[0], SpikeTrain))
            assert_same_attributes(block.segments[epoch_idx].spiketrains[0], sliced_st)

            self.assertTrue(isinstance(block.segments[epoch_idx].analogsignals[0], AnalogSignal))
            assert_same_attributes(block.segments[epoch_idx].analogsignals[0], sliced_anasig)

            if has_event:
                self.assertTrue(isinstance(block.segments[epoch_idx].events[0], Event))
                assert_same_attributes(block.segments[epoch_idx].events[0], sliced_event)

        block2 = Block()
        seg2 = Segment()
        epoch = Epoch(np.arange(10) * pq.s, durations=np.ones(10) * pq.s)
        epoch.annotate(pick="me instead")
        seg2.epochs = [proxy_epoch, epoch]
        block2.segments = [seg2]
        block2.check_relationships()

        # test correct loading and slicing of EpochProxy objects
        # (not tested above since we used the EpochProxy to cut the block)

        block3 = cut_block_by_epochs(block2, properties={"pick": "me instead"})

        for epoch_idx in range(len(epoch)):
            sliced_epoch = loaded_epoch.time_slice(
                t_start=epoch.times[epoch_idx], t_stop=epoch.times[epoch_idx] + epoch.durations[epoch_idx]
            )
            has_epoch = len(sliced_epoch) > 0

            if has_epoch:
                self.assertTrue(isinstance(block3.segments[epoch_idx].epochs[0], Epoch))
                assert_same_attributes(block3.segments[epoch_idx].epochs[0], sliced_epoch)


if __name__ == "__main__":
    unittest.main()



import unittest

from neo.rawio.openephysrawio import OpenEphysRawIO
from neo.test.rawiotest.common_rawio_test import BaseTestRawIO


class TestOpenEphysRawIO(
    BaseTestRawIO,
    unittest.TestCase,
):
    rawioclass = OpenEphysRawIO
    entities_to_download = ["openephys"]
    entities_to_test = [
        "openephys/OpenEphys_SampleData_1",
        # this file has gaps and this is now handle corretly
        "openephys/OpenEphys_SampleData_2_(multiple_starts)",
        # 'openephys/OpenEphys_SampleData_3',
        # two nodes with the new naming convention for openephys
        "openephys/openephys_rhythmdata_test_nodes/Record Node 120",
        "openephys/openephys_rhythmdata_test_nodes/Record Node 121",
    ]

    def test_raise_error_if_strange_timestamps(self):
        # In this dataset CH32 have strange timestamps
        reader = OpenEphysRawIO(dirname=self.get_local_path("openephys/OpenEphys_SampleData_3"))
        with self.assertRaises(Exception):
            reader.parse_header()

    def test_channel_order(self):
        reader = OpenEphysRawIO(dirname=self.get_local_path("openephys/OpenEphys_SampleData_1"))
        reader.parse_header()
        reader.header["signal_channels"]["name"][0].startswith("CH")


if __name__ == "__main__":
    unittest.main()



"""
RawIO for reading data from Neuralynx files.
This IO supports NCS, NEV, NSE and NTT file formats.


NCS contains the sampled signal for one channel
NEV contains events
NSE contains spikes and waveforms for mono electrodes
NTT contains spikes and waveforms for tetrodes
"""

from neo.rawio.neuralynxrawio.neuralynxrawio import NeuralynxRawIO



"""
Plexon2RawIO is a class to read Plexon PL2 files.

This IO is based on the pypl2lib module, see https://github.com/NeuralEnsemble/pypl2

On non-windows systems this IO requires Wine, see https://www.winehq.org/

Note on IO performance and memory consumption:
This IO is only partially lazy: To load a data chunk of an individual
signal channels, the complete data of that channel will be loaded into memory
and cached. This occurs as a limitation of the pypl2 package and might be improved
in future versions by exploiting the block-wise loading of signals.

To clear the currently cached signal data use the `clear_analogsignal_cache()` method.

Note on Neo header and spike channels:
The IO only considers enabled channels and will not list disabled channels in its header.

There is no 1-1 correspondence of PL2 spike channels and neo spike channels as each unit of
a PL2 spike channel will be represented as an individual neo spike channel.

Author: Julia Sprenger
"""

import pathlib
import warnings
import platform

from collections import namedtuple
from urllib.request import urlopen
from datetime import datetime

import numpy as np


from ..baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)


class Plexon2RawIO(BaseRawIO):
    """
    Class for "reading" data from a PL2 file

    Parameters
    ----------
    filename: str | Path
        The *.pl2 file to be loaded
    pl2_dll_file_path: str | Path | None, default: None
        The path to the necessary dll for loading pl2 files
        If None will find correct dll for architecture and if it does not exist will download it

    Notes
    -----
    * This IO is only partially lazy
    * The IO only considers enabled channels and will not list disabled channels in its header.
    * There is no 1-1 correspondence of PL2 spike channels and neo spike channels as each unit of
      a PL2 spike channel will be represented as an individual neo spike channel.

    Examples
    --------
    >>> import neo.rawio
    >>> r = neo.rawio.Plexon2RawIO(filename='my_data.pl2')
    >>> r.parse_header()
    >>> print(r)
    >>> raw_chunk = r.get_analogsignal_chunk(block_index=0,
                                             seg_index=0,
                                             i_start=0,
                                             i_stop=1024,
                                             stream_index=0,
                                             channel_indexes=range(10))
    >>> float_chunk = r.rescale_signal_raw_to_float(raw_chunk,
                                                    dtype='float64',
                                                    stream_index=0,
                                                    channel_indexes=[0, 3, 6])
    >>> spike_timestamp = r.get_spike_timestamps(spike_channel_index=0,
                                                 t_start=None,
                                                 t_stop=None)
    >>> spike_times = r.rescale_spike_timestamp(spike_timestamp, dtype='float64')
    >>> ev_timestamps, _, ev_labels = r.get_event_timestamps(event_channel_index=0)

    """

    extensions = ["pl2"]
    rawmode = "one-file"

    def __init__(self, filename, pl2_dll_file_path=None):

        # signals, event and spiking data will be cached
        # cached signal data can be cleared using `clear_analogsignal_cache()()`
        self._analogsignal_cache = {}
        self._event_channel_cache = {}
        self._spike_channel_cache = {}

        # note that this filename is used in self._source_name
        self.filename = pathlib.Path(filename)

        if (not self.filename.exists()) or (not self.filename.is_file()):
            raise ValueError(f"{self.filename} is not a file.")

        BaseRawIO.__init__(self)

        # download default PL2 dll once if not yet available
        if pl2_dll_file_path is None:
            architecture = platform.architecture()[0]
            if architecture == "64bit" and platform.system() in ["Windows", "Darwin"]:
                file_name = "PL2FileReader64.dll"
            else:  # Apparently wine uses the 32 bit version in linux
                file_name = "PL2FileReader.dll"
            pl2_dll_folder = pathlib.Path.home() / ".plexon_dlls_for_neo"
            pl2_dll_folder.mkdir(exist_ok=True)
            pl2_dll_file_path = pl2_dll_folder / file_name

            if not pl2_dll_file_path.exists():
                url = f"https://raw.githubusercontent.com/Neuralensemble/pypl2/master/bin/{file_name}"
                dist = urlopen(url=url)

                with open(pl2_dll_file_path, "wb") as f:
                    warnings.warn(f"dll file does not exist, downloading plexon dll to {pl2_dll_file_path}")
                    f.write(dist.read())

        # Instantiate wrapper for Windows DLL
        from neo.rawio.plexon2rawio.pypl2.pypl2lib import PyPL2FileReader

        self.pl2reader = PyPL2FileReader(pl2_dll_file_path=pl2_dll_file_path)

        reading_attempts = 10
        for attempt in range(reading_attempts):
            self.pl2reader.pl2_open_file(self.filename)

            # Verify the file handle is valid.
            if self.pl2reader._file_handle.value != 0:
                # File handle is valid, exit the loop early
                break
            else:
                if attempt == reading_attempts - 1:
                    self.pl2reader._print_error()
                    raise IOError(f"Opening {self.filename} failed after {reading_attempts} attempts.")

    def _source_name(self):
        return self.filename

    def _parse_header(self):
        """
        Collect information about the file, construct neo header and provide annotations
        """

        # Scanning sources and populating signal channels at the same time. Sources have to have
        # same sampling rate and number of samples to belong to one stream.
        signal_channels = []
        source_characteristics = {}
        Source = namedtuple("Source", "id name sampling_rate n_samples")
        for c in range(self.pl2reader.pl2_file_info.m_TotalNumberOfAnalogChannels):
            achannel_info = self.pl2reader.pl2_get_analog_channel_info(c)
            # only consider active channels
            if not achannel_info.m_ChannelEnabled:
                continue

            # assign to matching stream or create new stream based on signal characteristics
            rate = achannel_info.m_SamplesPerSecond
            n_samples = achannel_info.m_NumberOfValues
            source_id = str(achannel_info.m_Source)

            channel_source = Source(source_id, f"stream@{rate}Hz", rate, n_samples)
            existing_source = source_characteristics.setdefault(source_id, channel_source)

            # ensure that stream of this channel and existing stream have same properties
            if channel_source != existing_source:
                raise ValueError(
                    f"The channel source {channel_source} must be the same as the existing source {existing_source}"
                )

            ch_name = achannel_info.m_Name.decode()
            chan_id = f"source{achannel_info.m_Source}.{achannel_info.m_Channel}"
            dtype = "int16"
            units = achannel_info.m_Units.decode()
            gain = achannel_info.m_CoeffToConvertToUnits
            offset = 0.0  # PL2 files don't contain information on signal offset
            stream_id = source_id
            signal_channels.append((ch_name, chan_id, rate, dtype, units, gain, offset, stream_id))

        signal_channels = np.array(signal_channels, dtype=_signal_channel_dtype)
        self.signal_stream_characteristics = source_characteristics

        # create signal streams from source information
        signal_streams = []
        for stream_idx, source in source_characteristics.items():
            signal_streams.append((source.name, str(source.id)))
        signal_streams = np.array(signal_streams, dtype=_signal_stream_dtype)

        # pre-loading spike channel_data for later usage
        self._spike_channel_cache = {}
        spike_channels = []
        for c in range(self.pl2reader.pl2_file_info.m_TotalNumberOfSpikeChannels):
            schannel_info = self.pl2reader.pl2_get_spike_channel_info(c)

            # only consider active channels
            if not schannel_info.m_ChannelEnabled:
                continue

            for channel_unit_id in range(schannel_info.m_NumberOfUnits):
                unit_name = f"{schannel_info.m_Name.decode()}.{channel_unit_id}"
                unit_id = f"unit{schannel_info.m_Channel}.{channel_unit_id}"
                wf_units = schannel_info.m_Units
                wf_gain = schannel_info.m_CoeffToConvertToUnits
                wf_offset = 0.0  # A waveform offset is not provided in PL2 files
                wf_left_sweep = schannel_info.m_PreThresholdSamples
                wf_sampling_rate = schannel_info.m_SamplesPerSecond
                spike_channels.append(
                    (unit_name, unit_id, wf_units, wf_gain, wf_offset, wf_left_sweep, wf_sampling_rate)
                )

        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        # creating event/epoch channel
        self._event_channel_cache = {}
        event_channels = []
        for i in range(self.pl2reader.pl2_file_info.m_NumberOfDigitalChannels):
            echannel_info = self.pl2reader.pl2_get_digital_channel_info(i)

            # only consider active channels
            if not echannel_info.m_ChannelEnabled:
                continue

            # event channels are characterized by (name, id, type), with type in ['event', 'epoch']
            channel_name = echannel_info.m_Name.decode()
            event_channels.append((channel_name, echannel_info.m_Channel, "event"))

        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        # fill into header dict
        self.header = {}
        self.header["nb_block"] = 1
        self.header["nb_segment"] = [1]  # It seems pl2 can only contain a single segment
        self.header["signal_streams"] = signal_streams
        self.header["signal_channels"] = signal_channels
        self.header["spike_channels"] = spike_channels
        self.header["event_channels"] = event_channels

        self._generate_minimal_annotations()

        # this pprint lines really help for understand the nested (and complicated sometimes) dict
        # from pprint import pprint
        # pprint(self.raw_annotations)

        # Note: pl2_file_info.m_ReprocessorDateTime seems to be always empty.
        # To be checked against alternative pl2 reader.

        # Provide additional, recommended annotations for the final neo objects.
        block_index = 0
        bl_ann = self.raw_annotations["blocks"][block_index]
        bl_ann["name"] = f"Block containing PL2 data#{block_index}"
        bl_ann["file_origin"] = self.filename
        file_info = self.pl2reader.pl2_file_info
        block_info = {attr: getattr(file_info, attr) for attr, _ in file_info._fields_}

        # convert ctypes datetime objects to datetime.datetime objects for annotations
        from .pypl2.pypl2lib import tm

        for anno_key, anno_value in block_info.items():
            if isinstance(anno_value, tm):
                tmo = anno_value
                # invalid datetime information if year is <1
                if tmo.tm_year != 0:
                    microseconds = block_info["m_CreatorDateTimeMilliseconds"] * 1000
                    # tm_mon range is 0..11 https://cplusplus.com/reference/ctime/tm/
                    # python is 1..12 https://docs.python.org/3/library/datetime.html#datetime.datetime
                    # so month needs to be tm_mon+1; also tm_sec could cause problems in the case of leap
                    # seconds, but this is harder to defend against.
                    year = tmo.tm_year  # This has base 1900 in the c++ struct specification so we need to add 1900
                    year += 1900
                    dt = datetime(
                        year=year,
                        month=tmo.tm_mon + 1,
                        day=tmo.tm_mday,
                        hour=tmo.tm_hour,
                        minute=tmo.tm_min,
                        second=tmo.tm_sec,
                        microsecond=microseconds,
                    )
                    # ignoring daylight saving time information for now as timezone is unknown

                else:
                    dt = None

                block_info[anno_key] = dt

        bl_ann.update(block_info)
        for seg_index in range(1):
            seg_ann = bl_ann["segments"][seg_index]

            # some attributes don't apply to neo spike channels as these cover only a subpart of
            # the data of a PL2 spike channel
            spike_annotation_keys = [
                "m_SortEnabled",
                "m_SortRangeStart",
                "m_SortRangeEnd",
                "m_SourceTrodality",
                "m_OneBasedTrode",
                "m_OneBasedChannelInTrode",
                "m_Source",
                "m_Channel",
            ]
            for spike_channel_idx, spike_header in enumerate(self.header["spike_channels"]):
                schannel_name = spike_header["name"].split(".")[0]
                schannel_info = self.pl2reader.pl2_get_spike_channel_info_by_name(schannel_name)

                spiketrain_an = seg_ann["spikes"][spike_channel_idx]
                for key in spike_annotation_keys:
                    spiketrain_an[key] = getattr(schannel_info, key)

            event_annotation_keys = ["m_Source", "m_Channel", "m_Name"]
            for event_channel_idx, event_header in enumerate(self.header["event_channels"]):
                dchannel_name = event_header["name"]
                dchannel_info = self.pl2reader.pl2_get_digital_channel_info_by_name(dchannel_name)

                event_an = seg_ann["events"][event_channel_idx]
                for key in event_annotation_keys:
                    event_an[key] = getattr(dchannel_info, key)

            signal_array_annotation_keys = [
                "m_SourceTrodality",
                "m_OneBasedTrode",
                "m_OneBasedChannelInTrode",
                "m_Source",
                "m_Channel",
                "m_MaximumNumberOfFragments",
            ]

            for stream_idx, stream_header in enumerate(self.header["signal_streams"]):
                signal_array_annotations = {key: [] for key in signal_array_annotation_keys}
                stream_id = stream_header["id"]
                # extract values of individual signals

                stream_channel_mask = self.header["signal_channels"]["stream_id"] == stream_id
                for signal_idx, signal_header in enumerate(self.header["signal_channels"][stream_channel_mask]):
                    achannel_name = signal_header["name"]
                    achannel_info = self.pl2reader.pl2_get_analog_channel_info_by_name(achannel_name)
                    for key in signal_array_annotation_keys:
                        signal_array_annotations[key].append(getattr(achannel_info, key))

                seg_ann["signals"][stream_idx]["__array_annotations__"] = signal_array_annotations

    def _segment_t_start(self, block_index, seg_index):
        # this must return a float values in seconds
        return self.pl2reader.pl2_file_info.m_StartRecordingTime / self.pl2reader.pl2_file_info.m_TimestampFrequency

    def _segment_t_stop(self, block_index, seg_index):
        # this must return a float value in seconds
        end_time = (
            self.pl2reader.pl2_file_info.m_StartRecordingTime + self.pl2reader.pl2_file_info.m_DurationOfRecording
        )
        return end_time / self.pl2reader.pl2_file_info.m_TimestampFrequency

    def _get_signal_size(self, block_index, seg_index, stream_index):
        # this must return an integer value (the number of samples)

        stream_id = self.header["signal_streams"][stream_index]["id"]
        stream_characteristic = list(self.signal_stream_characteristics.values())[stream_index]
        if stream_id != stream_characteristic.id:
            raise ValueError(f"The `stream_id` must be {stream_characteristic.id}")
        return int(stream_characteristic.n_samples)  # Avoids returning a numpy.int64 scalar

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        # This returns the t_start of signals as a float value in seconds

        # TODO: Does the fragment_timestamp[0] need to be added here for digital signals?
        return self._segment_t_start(block_index, seg_index)

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index, channel_indexes):
        # this must return a signal chunk in a signal stream
        # limited with i_start/i_stop (can be None)
        # channel_indexes can be None (=all channel in the stream) or a list or numpy.array
        # This must return a numpy array 2D (even with one channel).
        # This must return the original dtype. No conversion here.
        # This must as fast as possible.
        # To speed up this call all preparatory calculations should be implemented
        # in _parse_header().

        stream_id = self.header["signal_streams"][stream_index]["id"]
        mask = self.header["signal_channels"]["stream_id"] == stream_id
        stream_channels = self.header["signal_channels"][mask]

        n_channels = len(stream_channels)
        n_samples = self.get_signal_size(block_index, seg_index, stream_index)

        if i_start is None:
            i_start = 0
        if i_stop is None:
            i_stop = n_samples

        if i_start < 0 or i_stop > n_samples:
            raise IndexError(f"Indexes ({i_start}, {i_stop}) out of range for signal with {n_samples} samples")

        # converting channel_indexes to array representation
        if channel_indexes is None:
            channel_indexes = np.arange(len(stream_channels), dtype="int")
        elif isinstance(channel_indexes, slice):
            channel_indexes = np.arange(len(stream_channels), dtype="int")[channel_indexes]
        else:
            channel_indexes = np.asarray(channel_indexes)

        # channel index sanity check
        if any(channel_indexes < 0) or any(channel_indexes >= n_channels):
            raise IndexError(f"Channel index out of range {channel_indexes} for stream with {n_channels} channels")

        nb_chan = len(channel_indexes)

        raw_signals = np.empty((i_stop - i_start, nb_chan), dtype="int16")
        for i, channel_idx in enumerate(channel_indexes):
            channel_name = stream_channels["name"][channel_idx]

            # use previously loaded channel data if possible
            if channel_name in self._analogsignal_cache:
                values = self._analogsignal_cache[channel_name]
            else:
                res = self.pl2reader.pl2_get_analog_channel_data_by_name(channel_name)
                fragment_timestamps, fragment_counts, values = res
                self._analogsignal_cache[channel_name] = values

            raw_signals[:, i] = values[i_start:i_stop]

        return raw_signals

    def clear_analogsignal_cache(self):
        for channel_name, values in self._analogsignal_cache.items():
            del values
        self._analogsignal_cache = {}

    def _spike_count(self, block_index, seg_index, spike_channel_index):
        channel_header = self.header["spike_channels"][spike_channel_index]
        channel_name, channel_unit_id = channel_header["name"].split(".")
        channel_unit_id = int(channel_unit_id)

        # loading spike channel data on demand when not already cached
        if channel_name not in self._spike_channel_cache:
            self._spike_channel_cache[channel_name] = self.pl2reader.pl2_get_spike_channel_data_by_name(channel_name)

        spike_timestamps, unit_ids, waveforms = self._spike_channel_cache[channel_name]
        nb_spikes = np.count_nonzero(unit_ids == channel_unit_id)

        return nb_spikes

    def _get_spike_timestamps(self, block_index, seg_index, spike_channel_index, t_start, t_stop):
        channel_header = self.header["spike_channels"][spike_channel_index]
        channel_name, channel_unit_id = channel_header["name"].split(".")
        channel_unit_id = int(channel_unit_id)

        # loading spike channel data on demand when not already cached
        if channel_name not in self._spike_channel_cache:
            self._spike_channel_cache[channel_name] = self.pl2reader.pl2_get_spike_channel_data_by_name(channel_name)

        spike_timestamps, unit_ids, waveforms = self._spike_channel_cache[channel_name]

        time_mask = self._get_timestamp_time_mask(t_start, t_stop, spike_timestamps)

        unit_mask = unit_ids[time_mask] == channel_unit_id
        spike_timestamps = spike_timestamps[time_mask][unit_mask]

        # spike timestamps are counted from the session start recording time
        spike_timestamps += self.pl2reader.pl2_file_info.m_StartRecordingTime

        return spike_timestamps

    def _rescale_spike_timestamp(self, spike_timestamps, dtype):
        # must rescale to second a particular spike_timestamps
        # with a fixed dtype so the user can choose the precision they want
        spike_times = spike_timestamps.astype(dtype)
        spike_times /= self.pl2reader.pl2_file_info.m_TimestampFrequency
        return spike_times

    def _get_spike_raw_waveforms(self, block_index, seg_index, spike_channel_index, t_start, t_stop):
        # this must return a 3D numpy array (nb_spike, nb_channel, nb_sample)
        # in the original dtype
        # this must be as fast as possible.
        # the same clip t_start/t_start must be used in _spike_timestamps()

        channel_header = self.header["spike_channels"][spike_channel_index]
        channel_name, channel_unit_id = channel_header["name"].split(".")

        # loading spike channel data on demand when not already cached
        if channel_name not in self._spike_channel_cache:
            self._spike_channel_cache[channel_name] = self.pl2reader.pl2_get_spike_channel_data_by_name(channel_name)

        spike_timestamps, unit_ids, waveforms = self._spike_channel_cache[channel_name]

        time_mask = self._get_timestamp_time_mask(t_start, t_stop, spike_timestamps)

        unit_mask = unit_ids[time_mask] == int(channel_unit_id)
        waveforms = waveforms[time_mask][unit_mask]

        # add tetrode dimension
        waveforms = np.expand_dims(waveforms, axis=1)
        return waveforms

    def _get_timestamp_time_mask(self, t_start, t_stop, timestamps):

        if t_start is not None or t_stop is not None:
            # restrict spikes to given limits (in seconds)
            timestamp_frequency = self.pl2reader.pl2_file_info.m_TimestampFrequency
            lim0 = int(t_start * timestamp_frequency)
            lim1 = int(t_stop * self.pl2reader.pl2_file_info.m_TimestampFrequency)

            # limits are with respect to segment t_start and not to time 0
            lim0 -= self.pl2reader.pl2_file_info.m_StartRecordingTime
            lim1 -= self.pl2reader.pl2_file_info.m_StartRecordingTime

            time_mask = (timestamps >= lim0) & (timestamps <= lim1)

        else:
            time_mask = slice(None, None)

        return time_mask

    def _event_count(self, block_index, seg_index, event_channel_index):

        channel_header = self.header["event_channels"][event_channel_index]
        channel_name = channel_header["name"]

        # loading event channel data on demand when not already cached
        if channel_name not in self._event_channel_cache:
            self._event_channel_cache[channel_name] = self.pl2reader.pl2_get_digital_channel_data_by_name(channel_name)

        event_times, values = self._event_channel_cache[channel_name]

        return len(event_times)

    def _get_event_timestamps(self, block_index, seg_index, event_channel_index, t_start, t_stop):
        # the main difference between spike channel and event channel
        # is that for here we have 3 numpy array timestamp, durations, labels
        # durations must be None for 'event'
        # label must a dtype ='U'

        channel_header = self.header["event_channels"][event_channel_index]
        channel_name = channel_header["name"]

        # loading event channel data on demand when not already cached
        if channel_name not in self._event_channel_cache:
            self._event_channel_cache[channel_name] = self.pl2reader.pl2_get_digital_channel_data_by_name(channel_name)

        event_times, labels = self._event_channel_cache[channel_name]
        labels = np.asarray(labels, dtype="U")

        time_mask = self._get_timestamp_time_mask(t_start, t_stop, event_times)

        # events don't have a duration. Epochs are not supported
        durations = None

        # event timestamps are counted from the session start recording time
        return_times = event_times[time_mask] + self.pl2reader.pl2_file_info.m_StartRecordingTime

        return return_times, durations, labels[time_mask]

    def _rescale_event_timestamp(self, event_timestamps, dtype, event_channel_index):
        # must rescale to second a particular event_timestamps
        # with a fixed dtype so the user can choose the precision he want.

        event_times = event_timestamps.astype(dtype)
        event_times /= self.pl2reader.pl2_file_info.m_TimestampFrequency
        return event_times

    def _rescale_epoch_duration(self, raw_duration, dtype, event_channel_index):
        durations = raw_duration.astype(dtype)
        durations /= self.pl2reader.pl2_file_info.m_TimestampFrequency
        return durations

    def close(self):
        self.pl2reader.pl2_close_all_files()



"""
Tools for IO coder:
  * Creating RecordingChannel and making links with AnalogSignals and
    SPikeTrains
"""

from collections.abc import MutableSequence

from neo.core import (
    AnalogSignal,
    Block,
    Epoch,
    Event,
    IrregularlySampledSignal,
    Group,
    ChannelView,
    Segment,
    SpikeTrain,
)


class LazyList(MutableSequence):
    """An enhanced list that can load its members on demand. Behaves exactly
    like a regular list for members that are Neo objects. Each item should
    contain the information that ``load_lazy_cascade`` needs to load the
    respective object.
    """

    _container_objects = {Block, Segment, Group}
    _neo_objects = _container_objects.union(
        [AnalogSignal, Epoch, Event, ChannelView, IrregularlySampledSignal, SpikeTrain]
    )

    def __init__(self, io, lazy, items=None):
        """
        :param io: IO instance that can load items.
        :param lazy: Lazy parameter with which the container object
            using the list was loaded.
        :param items: Optional, initial list of items.
        """
        if items is None:
            self._data = []
        else:
            self._data = items
        self._lazy = lazy
        self._io = io

    def __getitem__(self, index):
        item = self._data.__getitem__(index)
        if isinstance(index, slice):
            return LazyList(self._io, item)

        if type(item) in self._neo_objects:
            return item

        loaded = self._io.load_lazy_cascade(item, self._lazy)
        self._data[index] = loaded
        return loaded

    def __delitem__(self, index):
        self._data.__delitem__(index)

    def __len__(self):
        return self._data.__len__()

    def __setitem__(self, index, value):
        self._data.__setitem__(index, value)

    def insert(self, index, value):
        self._data.insert(index, value)

    def append(self, value):
        self._data.append(value)

    def reverse(self):
        self._data.reverse()

    def extend(self, values):
        self._data.extend(values)

    def remove(self, value):
        self._data.remove(value)

    def __str__(self):
        return "<" + self.__class__.__name__ + ">" + self._data.__str__()

    def __repr__(self):
        return "<" + self.__class__.__name__ + ">" + self._data.__repr__()



from neo.io.basefromrawio import BaseFromRaw
from neo.rawio.rawmcsrawio import RawMCSRawIO


class RawMCSIO(RawMCSRawIO, BaseFromRaw):
    _prefered_signal_group_mode = "group-by-same-units"

    def __init__(self, filename):
        RawMCSRawIO.__init__(self, filename=filename)
        BaseFromRaw.__init__(self, filename)



"""
Tests of the neo.core.segment.Segment class
"""

from copy import deepcopy

from datetime import datetime

import unittest

import numpy as np
import quantities as pq

try:
    from IPython.lib.pretty import pretty
except ImportError as err:
    HAVE_IPYTHON = False
else:
    HAVE_IPYTHON = True

from neo.core.segment import Segment
from neo.core import AnalogSignal, Block, Event, IrregularlySampledSignal, Epoch, SpikeTrain
from neo.core.spiketrainlist import SpikeTrainList
from neo.core.container import filterdata
from neo.test.tools import assert_neo_object_is_compliant, assert_same_sub_schema, assert_same_attributes
from neo.test.generate_datasets import random_segment, simple_block
from neo.rawio.examplerawio import ExampleRawIO
from neo.io.proxyobjects import AnalogSignalProxy, SpikeTrainProxy, EventProxy, EpochProxy


N_EXAMPLES = 5


class TestSegment(unittest.TestCase):

    def setUp(self):
        self.segments = [random_segment() for i in range(N_EXAMPLES)]

    def test_init(self):
        seg = Segment(name="a segment", index=3)
        assert_neo_object_is_compliant(seg)
        self.assertEqual(seg.name, "a segment")
        self.assertEqual(seg.file_origin, None)
        self.assertEqual(seg.index, 3)

    def test_times(self):
        for seg in self.segments:
            # calculate target values for t_start and t_stop
            t_starts, t_stops = [], []
            for children in [seg.analogsignals, seg.epochs, seg.events, seg.irregularlysampledsignals, seg.spiketrains]:
                for child in children:
                    if hasattr(child, "t_start"):
                        t_starts.append(child.t_start)
                    if hasattr(child, "t_stop"):
                        t_stops.append(child.t_stop)
                    if hasattr(child, "time"):
                        t_starts.append(child.time)
                        t_stops.append(child.time)
                    if hasattr(child, "times"):
                        t_starts.append(child.times[0])
                        t_stops.append(child.times[-1])
            targ_t_start = min(t_starts)
            targ_t_stop = max(t_stops)

            self.assertEqual(seg.t_start, targ_t_start)
            self.assertEqual(seg.t_stop, targ_t_stop)

        # Testing times with ProxyObjects
        seg = Segment()
        reader = ExampleRawIO(filename="my_filename.fake")
        reader.parse_header()

        proxy_anasig = AnalogSignalProxy(
            rawio=reader, stream_index=0, inner_stream_channels=None, block_index=0, seg_index=0
        )
        seg.analogsignals.append(proxy_anasig)

        proxy_st = SpikeTrainProxy(rawio=reader, spike_channel_index=0, block_index=0, seg_index=0)
        seg.spiketrains.append(proxy_st)

        proxy_event = EventProxy(rawio=reader, event_channel_index=0, block_index=0, seg_index=0)
        seg.events.append(proxy_event)

        proxy_epoch = EpochProxy(rawio=reader, event_channel_index=1, block_index=0, seg_index=0)
        seg.epochs.append(proxy_epoch)

        t_starts, t_stops = [], []
        for children in [seg.analogsignals, seg.epochs, seg.events, seg.irregularlysampledsignals, seg.spiketrains]:
            for child in children:
                if hasattr(child, "t_start"):
                    t_starts.append(child.t_start)
                if hasattr(child, "t_stop"):
                    t_stops.append(child.t_stop)
                if hasattr(child, "time"):
                    t_starts.append(child.time)
                    t_stops.append(child.time)
                if hasattr(child, "times"):
                    t_starts.append(child.times[0])
                    t_stops.append(child.times[-1])
        targ_t_start = min(t_starts)
        targ_t_stop = max(t_stops)

        self.assertEqual(seg.t_start, targ_t_start)
        self.assertEqual(seg.t_stop, targ_t_stop)

    def test__merge(self):
        seg1 = self.segments[0]
        seg2 = self.segments[1]
        orig_seg1 = deepcopy(seg1)
        seg1.merge(seg2)

        assert_same_sub_schema(orig_seg1.analogsignals + seg2.analogsignals, seg1.analogsignals)
        assert_same_sub_schema(
            orig_seg1.irregularlysampledsignals + seg2.irregularlysampledsignals, seg1.irregularlysampledsignals
        )
        assert_same_sub_schema(orig_seg1.epochs + seg2.epochs, seg1.epochs)
        assert_same_sub_schema(orig_seg1.events + seg2.events, seg1.events)
        assert_same_sub_schema(orig_seg1.spiketrains + seg2.spiketrains, seg1.spiketrains)

    def test__size(self):
        for segment in self.segments:
            targ1 = {
                "epochs": len(segment.epochs),
                "events": len(segment.events),
                "irregularlysampledsignals": len(segment.irregularlysampledsignals),
                "spiketrains": len(segment.spiketrains),
                "analogsignals": len(segment.analogsignals),
                "imagesequences": len(segment.imagesequences),
            }
            self.assertEqual(segment.size, targ1)

    def test__filter_none(self):
        for segment in self.segments:
            targ = []
            # collecting all data objects in target block
            targ.extend(segment.analogsignals)
            targ.extend(segment.epochs)
            targ.extend(segment.events)
            targ.extend(segment.irregularlysampledsignals)
            targ.extend(segment.spiketrains)
            targ.extend(segment.imagesequences)

            # occasionally we randomly get only spike trains,
            # and then we have to convert to a SpikeTrainList
            # to match the output of segment.filter
            if all(isinstance(obj, SpikeTrain) for obj in targ):
                targ = SpikeTrainList(items=targ, parent=segment)

            res0 = segment.filter()
            res1 = segment.filter({})
            res2 = segment.filter([])
            res3 = segment.filter([{}])
            res4 = segment.filter([{}, {}])
            res5 = segment.filter([{}, {}])
            res6 = segment.filter(targdict={})
            res7 = segment.filter(targdict=[])
            res8 = segment.filter(targdict=[{}])
            res9 = segment.filter(targdict=[{}, {}])

            assert_same_sub_schema(res0, targ)
            assert_same_sub_schema(res1, targ)
            assert_same_sub_schema(res2, targ)
            assert_same_sub_schema(res3, targ)
            assert_same_sub_schema(res4, targ)
            assert_same_sub_schema(res5, targ)
            assert_same_sub_schema(res6, targ)
            assert_same_sub_schema(res7, targ)
            assert_same_sub_schema(res8, targ)
            assert_same_sub_schema(res9, targ)

    def test__filter_annotation_single(self):
        segment = simple_block().segments[0]
        targ = [segment.analogsignals[0], segment.spiketrains[1]]

        res0 = segment.filter(thing="wotsit")
        res1 = segment.filter({"thing": "wotsit"})
        res2 = segment.filter(targdict={"thing": "wotsit"})
        res3 = segment.filter([{"thing": "wotsit"}])
        res4 = segment.filter(targdict=[{"thing": "wotsit"}])

        assert_same_sub_schema(res0, targ)
        assert_same_sub_schema(res1, targ)
        assert_same_sub_schema(res2, targ)
        assert_same_sub_schema(res3, targ)
        assert_same_sub_schema(res4, targ)

    def test__filter_single_annotation_nores(self):
        segment = simple_block().segments[0]
        targ = []

        res0 = segment.filter(j=5)
        res1 = segment.filter({"j": 5})
        res2 = segment.filter(targdict={"j": 5})
        res3 = segment.filter([{"j": 5}])
        res4 = segment.filter(targdict=[{"j": 5}])

        assert_same_sub_schema(res0, targ)
        assert_same_sub_schema(res1, targ)
        assert_same_sub_schema(res2, targ)
        assert_same_sub_schema(res3, targ)
        assert_same_sub_schema(res4, targ)

    def test__filter_attribute_single(self):
        segment = simple_block().segments[1]

        targ = [segment.analogsignals[0], segment.irregularlysampledsignals[0]]

        name = targ[0].name
        res0 = segment.filter(name=name)
        res1 = segment.filter({"name": name})
        res2 = segment.filter(targdict={"name": name})

        assert_same_sub_schema(res0, targ)
        assert_same_sub_schema(res1, targ)
        assert_same_sub_schema(res2, targ)

    def test__filter_attribute_single_nores(self):
        segment = simple_block().segments[1]

        name = "potato"
        res0 = segment.filter(name=name)
        res1 = segment.filter({"name": name})
        res2 = segment.filter(targdict={"name": name})

        self.assertEqual(len(res0), 0)
        self.assertEqual(len(res1), 0)
        self.assertEqual(len(res2), 0)

    def test__filter_multi(self):

        segment = simple_block().segments[1]
        targ = [segment.analogsignals[0], segment.irregularlysampledsignals[0]]

        filter = {"name": targ[0].name, "thing": targ[0].annotations["thing"]}

        res0 = segment.filter(**filter)
        res1 = segment.filter(filter)
        res2 = segment.filter(targdict=filter)

        assert_same_sub_schema(res0, targ)
        assert_same_sub_schema(res1, targ)
        assert_same_sub_schema(res2, targ)

    def test__filter_multi_nores(self):
        segment = simple_block().segments[0]
        targ = []

        filter = {"name": "carrot", "thing": "another thing"}

        res0 = segment.filter(**filter)
        res1 = segment.filter(filter)
        res2 = segment.filter(targdict=filter)

        assert_same_sub_schema(res0, targ)
        assert_same_sub_schema(res1, targ)
        assert_same_sub_schema(res2, targ)

    def test__filter_multi_partres(self):
        segment = simple_block().segments[1]
        targ = [segment.analogsignals[0], segment.irregularlysampledsignals[0]]

        filter = {"name": targ[0].name, "thing": "another thing"}

        res0 = segment.filter(**filter)
        res1 = segment.filter(filter)
        res2 = segment.filter(targdict=filter)

        assert_same_sub_schema(res0, targ)
        assert_same_sub_schema(res1, targ)
        assert_same_sub_schema(res2, targ)

    def test__filter_no_annotation_but_object(self):
        for segment in self.segments:
            targ = segment.spiketrains
            assert isinstance(targ, SpikeTrainList)
            res = segment.filter(objects=SpikeTrain)
            if len(res) > 0:
                # if res has length 0 it will be just a plain list
                assert isinstance(res, SpikeTrainList)
                assert_same_sub_schema(res, targ)

            targ = segment.analogsignals
            res = segment.filter(objects=AnalogSignal)
            assert_same_sub_schema(res, targ)

            targ = segment.analogsignals + segment.spiketrains
            res = segment.filter(objects=[AnalogSignal, SpikeTrain])
            if len(res) > 0:
                assert_same_sub_schema(res, targ)

    def test__filter_single_annotation_obj_single(self):
        segment = simple_block().segments[0]
        targ = [segment.analogsignals[1]]

        res0 = segment.filter(thing="frooble", objects="AnalogSignal")
        res1 = segment.filter(thing="frooble", objects=AnalogSignal)
        res2 = segment.filter(thing="frooble", objects=["AnalogSignal"])
        res3 = segment.filter(thing="frooble", objects=[AnalogSignal])

        assert_same_sub_schema(res0, targ)
        assert_same_sub_schema(res1, targ)
        assert_same_sub_schema(res2, targ)
        assert_same_sub_schema(res3, targ)

    def test__filter_single_annotation_obj_multi(self):
        segment = simple_block().segments[0]
        targ = [segment.analogsignals[0], segment.spiketrains[1]]

        res0 = segment.filter(thing="wotsit", objects=["Event", SpikeTrain, "AnalogSignal"])

        assert_same_sub_schema(res0, targ)

    def test__filter_single_annotation_obj_none(self):
        segment = simple_block().segments[0]
        targ = []

        res0 = segment.filter(j=1, objects=Epoch)
        res1 = segment.filter(j=1, objects="Epoch")
        res2 = segment.filter(j=1, objects=[])

        assert_same_sub_schema(res0, targ)
        assert_same_sub_schema(res1, targ)
        assert_same_sub_schema(res2, targ)

    def test__filter_single_annotation_nodata(self):
        segment = simple_block().segments[0]
        targ = []
        res0 = segment.filter(thing="frooble", data=False)
        assert_same_sub_schema(res0, targ)

    def test__filter_single_attribute_nodata(self):
        segment = simple_block().segments[0]
        targ = []
        res0 = segment.filter(name=segment.analogsignals[0], data=False)
        assert_same_sub_schema(res0, targ)

    def test__filter_single_annotation_container(self):
        segment = simple_block().segments[1]

        targ = [
            segment.analogsignals[0],
        ]

        res0 = segment.filter(thing="amajig", container=True)

        self.assertEqual(res0, targ)
        assert_same_sub_schema(res0, targ)

    def test__filter_single_attribute_container(self):
        segment = simple_block().segments[1]

        targ = [segment.analogsignals[0], segment.irregularlysampledsignals[0]]
        res0 = segment.filter(name=targ[0].name, container=True)

        self.assertEqual(res0, targ)
        assert_same_sub_schema(res0, targ)

    def test__filter_single_annotation_container(self):
        segment = simple_block().segments[1]

        targ = []
        res0 = segment.filter(thing="amajig", container=True, data=False)

        self.assertEqual(res0, targ)
        assert_same_sub_schema(res0, targ)

    def test__filter_single_attribute_nodata_container(self):
        segment = simple_block().segments[1]

        targ = []
        res0 = segment.filter(name=segment.analogsignals[0].name, container=True, data=False)

        self.assertEqual(res0, targ)
        assert_same_sub_schema(res0, targ)

    # @unittest.skipUnless(HAVE_IPYTHON, "requires IPython")
    # def test__pretty(self):
    #     ann = get_annotations()
    #     ann['seed'] = self.seed1
    #     ann = pretty(ann).replace('\n ', '\n  ')
    #     res = pretty(self.seg1)
    #
    #     sigarr0 = pretty(self.sigarrs1[0])
    #     sigarr1 = pretty(self.sigarrs1[1])
    #     sigarr0 = sigarr0.replace('\n', '\n   ')
    #     sigarr1 = sigarr1.replace('\n', '\n   ')
    #
    #     targ = ("Segment with " +
    #             ("%s analogsignals, " %
    #              (len(self.sigarrs1a),)) +
    #             ("%s epochs, " % len(self.epcs1a)) +
    #             ("%s events, " % len(self.evts1a)) +
    #             ("%s irregularlysampledsignals, " %
    #              len(self.irsigs1a)) +
    #             ("%s spiketrains\n" % len(self.trains1a)) +
    #             ("name: '%s'\ndescription: '%s'\n" %
    #              (self.seg1.name, self.seg1.description)
    #              ) +
    #
    #             ("annotations: %s\n" % ann) +
    #
    #             ("# analogsignals (N=%s)\n" % len(self.sigarrs1a)) +
    #
    #             ('%s: %s\n' % (0, sigarr0)) +
    #             ('%s: %s' % (1, sigarr1)))
    #
    #     self.assertEqual(res, targ)

    def test__time_slice(self):
        time_slice = [0.5, 5.6] * pq.s

        epoch2 = Epoch([0.6, 9.5, 16.8, 34.1] * pq.s, durations=[4.5, 4.8, 5.0, 5.0] * pq.s, t_start=0.1 * pq.s)
        epoch2.annotate(epoch_type="b")
        epoch2.array_annotate(trial_id=[1, 2, 3, 4])

        event = Event(times=[0.5, 10.0, 25.2] * pq.s, t_start=0.1 * pq.s)
        event.annotate(event_type="trial start")
        event.array_annotate(trial_id=[1, 2, 3])

        anasig = AnalogSignal(np.arange(50.0) * pq.mV, t_start=0.1 * pq.s, sampling_rate=1.0 * pq.Hz)
        irrsig = IrregularlySampledSignal(signal=np.arange(50.0) * pq.mV, times=anasig.times, t_start=0.1 * pq.s)
        st = SpikeTrain(
            np.arange(0.5, 50, 7) * pq.s,
            t_start=0.1 * pq.s,
            t_stop=50.0 * pq.s,
            waveforms=np.array(
                [
                    [[0.0, 1.0], [0.1, 1.1]],
                    [[2.0, 3.0], [2.1, 3.1]],
                    [[4.0, 5.0], [4.1, 5.1]],
                    [[6.0, 7.0], [6.1, 7.1]],
                    [[8.0, 9.0], [8.1, 9.1]],
                    [[12.0, 13.0], [12.1, 13.1]],
                    [[14.0, 15.0], [14.1, 15.1]],
                    [[16.0, 17.0], [16.1, 17.1]],
                ]
            )
            * pq.mV,
            array_annotations={"spikenum": np.arange(1, 9)},
        )

        seg = Segment()
        seg.epochs = [epoch2]
        seg.events = [event]
        seg.analogsignals = [anasig]
        seg.irregularlysampledsignals = [irrsig]
        seg.spiketrains = [st]

        block = Block()
        block.segments = [seg]
        block.check_relationships()

        # test without resetting the time
        sliced = seg.time_slice(time_slice[0], time_slice[1])

        assert_neo_object_is_compliant(sliced)

        self.assertEqual(len(sliced.events), 1)
        self.assertEqual(len(sliced.spiketrains), 1)
        self.assertEqual(len(sliced.analogsignals), 1)
        self.assertEqual(len(sliced.irregularlysampledsignals), 1)
        self.assertEqual(len(sliced.epochs), 1)

        assert_same_attributes(sliced.spiketrains[0], st.time_slice(t_start=time_slice[0], t_stop=time_slice[1]))
        assert_same_attributes(sliced.analogsignals[0], anasig.time_slice(t_start=time_slice[0], t_stop=time_slice[1]))
        assert_same_attributes(
            sliced.irregularlysampledsignals[0], irrsig.time_slice(t_start=time_slice[0], t_stop=time_slice[1])
        )
        assert_same_attributes(sliced.events[0], event.time_slice(t_start=time_slice[0], t_stop=time_slice[1]))
        assert_same_attributes(sliced.epochs[0], epoch2.time_slice(t_start=time_slice[0], t_stop=time_slice[1]))

        seg = Segment()
        seg.epochs = [epoch2]
        seg.events = [event]
        seg.analogsignals = [anasig]
        seg.irregularlysampledsignals = [irrsig]
        seg.spiketrains = [st]

        block = Block()
        block.segments = [seg]
        block.check_relationships()

        # test with resetting the time
        sliced = seg.time_slice(time_slice[0], time_slice[1], reset_time=True)

        assert_neo_object_is_compliant(sliced)

        self.assertEqual(len(sliced.events), 1)
        self.assertEqual(len(sliced.spiketrains), 1)
        self.assertEqual(len(sliced.analogsignals), 1)
        self.assertEqual(len(sliced.irregularlysampledsignals), 1)
        self.assertEqual(len(sliced.epochs), 1)

        assert_same_attributes(
            sliced.spiketrains[0],
            st.time_shift(-time_slice[0]).time_slice(t_start=0 * pq.s, t_stop=time_slice[1] - time_slice[0]),
        )

        anasig_target = anasig.copy()
        anasig_target = anasig_target.time_shift(-time_slice[0]).time_slice(
            t_start=0 * pq.s, t_stop=time_slice[1] - time_slice[0]
        )
        assert_same_attributes(sliced.analogsignals[0], anasig_target)
        irrsig_target = irrsig.copy()
        irrsig_target = irrsig_target.time_shift(-time_slice[0]).time_slice(
            t_start=0 * pq.s, t_stop=time_slice[1] - time_slice[0]
        )
        assert_same_attributes(sliced.irregularlysampledsignals[0], irrsig_target)
        assert_same_attributes(
            sliced.events[0],
            event.time_shift(-time_slice[0]).time_slice(t_start=0 * pq.s, t_stop=time_slice[1] - time_slice[0]),
        )
        assert_same_attributes(
            sliced.epochs[0],
            epoch2.time_shift(-time_slice[0]).time_slice(t_start=0 * pq.s, t_stop=time_slice[1] - time_slice[0]),
        )

        seg = Segment()

        reader = ExampleRawIO(filename="my_filename.fake")
        reader.parse_header()

        proxy_anasig = AnalogSignalProxy(
            rawio=reader, stream_index=0, inner_stream_channels=None, block_index=0, seg_index=0
        )
        seg.analogsignals.append(proxy_anasig)

        proxy_st = SpikeTrainProxy(rawio=reader, spike_channel_index=0, block_index=0, seg_index=0)
        seg.spiketrains.append(proxy_st)

        proxy_event = EventProxy(rawio=reader, event_channel_index=0, block_index=0, seg_index=0)
        seg.events.append(proxy_event)

        proxy_epoch = EpochProxy(rawio=reader, event_channel_index=1, block_index=0, seg_index=0)
        proxy_epoch.annotate(pick="me")
        seg.epochs.append(proxy_epoch)

        loaded_epoch = proxy_epoch.load()
        loaded_event = proxy_event.load()
        loaded_st = proxy_st.load()
        loaded_anasig = proxy_anasig.load()

        block = Block()
        block.segments = [seg]
        block.check_relationships()

        # test with proxy objects
        sliced = seg.time_slice(time_slice[0], time_slice[1])

        assert_neo_object_is_compliant(sliced)

        sliced_event = loaded_event.time_slice(t_start=time_slice[0], t_stop=time_slice[1])
        has_event = len(sliced_event) > 0

        sliced_anasig = loaded_anasig.time_slice(t_start=time_slice[0], t_stop=time_slice[1])

        sliced_st = loaded_st.time_slice(t_start=time_slice[0], t_stop=time_slice[1])

        self.assertEqual(len(sliced.events), int(has_event))
        self.assertEqual(len(sliced.spiketrains), 1)
        self.assertEqual(len(sliced.analogsignals), 1)

        self.assertTrue(isinstance(sliced.spiketrains[0], SpikeTrain))
        assert_same_attributes(sliced.spiketrains[0], sliced_st)

        self.assertTrue(isinstance(sliced.analogsignals[0], AnalogSignal))
        assert_same_attributes(sliced.analogsignals[0], sliced_anasig)

        if has_event:
            self.assertTrue(isinstance(sliced.events[0], Event))
            assert_same_attributes(sliced.events[0], sliced_event)

    def test_time_slice_None(self):
        time_slices = [(None, 5.0 * pq.s), (5.0 * pq.s, None), (None, None)]

        anasig = AnalogSignal(np.arange(50.0) * pq.mV, sampling_rate=1.0 * pq.Hz)
        seg = Segment()
        seg.analogsignals = [anasig]

        block = Block()
        block.segments = [seg]
        block.check_relationships()

        # test without resetting the time
        for t_start, t_stop in time_slices:
            sliced = seg.time_slice(t_start, t_stop)

            assert_neo_object_is_compliant(sliced)
            self.assertEqual(len(sliced.analogsignals), 1)

            exp_t_start, exp_t_stop = t_start, t_stop
            if exp_t_start is None:
                exp_t_start = seg.t_start
            if exp_t_stop is None:
                exp_t_stop = seg.t_stop

            self.assertEqual(exp_t_start, sliced.t_start)
            self.assertEqual(exp_t_stop, sliced.t_stop)

    def test__deepcopy(self):
        childconts = ("analogsignals", "epochs", "events", "irregularlysampledsignals", "spiketrains")
        for segment in self.segments:
            seg1_copy = deepcopy(segment)

            # Same structure top-down, i.e. links from parents to children are correct
            assert_same_sub_schema(seg1_copy, segment)

            # Correct structure bottom-up, i.e. links from children to parents are correct
            # No need to cascade, all children are leaves, i.e. don't have any children
            for childtype in childconts:
                for child in getattr(seg1_copy, childtype, []):
                    self.assertEqual(id(child.segment), id(seg1_copy))

    def test_add(self):
        seg = Segment()

        reader = ExampleRawIO(filename="my_filename.fake")
        reader.parse_header()

        proxy_anasig = AnalogSignalProxy(
            rawio=reader, stream_index=0, inner_stream_channels=None, block_index=0, seg_index=0
        )
        seg.add(proxy_anasig)
        assert len(seg.analogsignals) == 1

        proxy_st = SpikeTrainProxy(rawio=reader, spike_channel_index=0, block_index=0, seg_index=0)
        seg.add(proxy_st)
        assert len(seg.spiketrains) == 1

        proxy_event = EventProxy(rawio=reader, event_channel_index=0, block_index=0, seg_index=0)
        seg.add(proxy_event)
        assert len(seg.events) == 1

        proxy_epoch = EpochProxy(rawio=reader, event_channel_index=1, block_index=0, seg_index=0)
        seg.add(proxy_epoch)
        assert len(seg.epochs) == 1

    def test_add_invalid_type_raises_Exception(self):
        seg = Segment()
        self.assertRaises(TypeError, seg.add, Block())


if __name__ == "__main__":
    unittest.main()



"""
This module implements :class:`Group`, which represents a subset of the
channels in an :class:`AnalogSignal` or :class:`IrregularlySampledSignal`.

It replaces and extends the grouping function of the former :class:`ChannelIndex`
and :class:`Unit`.
"""

from neo.core.container import Container
from neo.core.analogsignal import AnalogSignal
from neo.core.container import Container
from neo.core.objectlist import ObjectList
from neo.core.epoch import Epoch
from neo.core.event import Event
from neo.core.imagesequence import ImageSequence
from neo.core.irregularlysampledsignal import IrregularlySampledSignal
from neo.core.segment import Segment
from neo.core.spiketrainlist import SpikeTrainList
from neo.core.view import ChannelView
from neo.core.regionofinterest import RegionOfInterest


class Group(Container):
    """
    Can contain any of the data objects, views, or other groups,
    outside the hierarchy of the segment and block containers.
    A common use is to link the :class:`SpikeTrain` objects within a :class:`Block`,
    possibly across multiple Segments, that were emitted by the same neuron.

    *Required attributes/properties*:
        None

    *Recommended attributes/properties*:
        :objects: (Neo object) Objects with which to pre-populate the :class:`Group`
        :name: (str) A label for the group.
        :description: (str) Text description.
        :file_origin: (str) Filesystem path or URL of the original data file.

    *Optional arguments*:
        :allowed_types: (list or tuple) Types of Neo object that are allowed to be
          added to the Group. If not specified, any Neo object can be added.

    Note: Any other additional arguments are assumed to be user-specific
            metadata and stored in :attr:`annotations`.

    *Container of*:
        :class:`AnalogSignal`, :class:`IrregularlySampledSignal`, :class:`SpikeTrain`,
        :class:`Event`, :class:`Epoch`, :class:`ChannelView`, :class:`Group`
    """

    _data_child_objects = (
        "AnalogSignal",
        "IrregularlySampledSignal",
        "SpikeTrain",
        "Event",
        "Epoch",
        "ChannelView",
        "ImageSequence",
        "CircularRegionOfInterest",
        "RectangularRegionOfInterest",
        "PolygonRegionOfInterest",
    )
    _container_child_objects = ("Group",)
    _parent_objects = ("Block",)

    def __init__(self, objects=None, name=None, description=None, file_origin=None, allowed_types=None, **annotations):
        super().__init__(name=name, description=description, file_origin=file_origin, **annotations)

        # note that we create the ObjectLists here _without_ a parent argument
        # since objects do not have a reference to the group(s)
        # they are contained in.
        self._analogsignals = ObjectList(AnalogSignal)
        self._irregularlysampledsignals = ObjectList(IrregularlySampledSignal)
        self._spiketrains = SpikeTrainList()
        self._events = ObjectList(Event)
        self._epochs = ObjectList(Epoch)
        self._channelviews = ObjectList(ChannelView)
        self._imagesequences = ObjectList(ImageSequence)
        self._regionsofinterest = ObjectList(RegionOfInterest)
        self._segments = ObjectList(Segment)  # to remove?
        self._groups = ObjectList(Group)

        if allowed_types is None:
            self.allowed_types = None
        else:
            self.allowed_types = tuple(allowed_types)
            for type_ in self.allowed_types:
                if type_.__name__ not in self._child_objects:
                    raise TypeError(f"Groups can not contain objects of type {type_.__name__}")

        if objects:
            self.add(*objects)

    analogsignals = property(
        fget=lambda self: self._get_object_list("_analogsignals"),
        fset=lambda self, value: self._set_object_list("_analogsignals", value),
        doc="list of AnalogSignals contained in this group",
    )

    irregularlysampledsignals = property(
        fget=lambda self: self._get_object_list("_irregularlysampledsignals"),
        fset=lambda self, value: self._set_object_list("_irregularlysampledsignals", value),
        doc="list of IrregularlySignals contained in this group",
    )

    events = property(
        fget=lambda self: self._get_object_list("_events"),
        fset=lambda self, value: self._set_object_list("_events", value),
        doc="list of Events contained in this group",
    )

    epochs = property(
        fget=lambda self: self._get_object_list("_epochs"),
        fset=lambda self, value: self._set_object_list("_epochs", value),
        doc="list of Epochs contained in this group",
    )

    channelviews = property(
        fget=lambda self: self._get_object_list("_channelviews"),
        fset=lambda self, value: self._set_object_list("_channelviews", value),
        doc="list of ChannelViews contained in this group",
    )

    imagesequences = property(
        fget=lambda self: self._get_object_list("_imagesequences"),
        fset=lambda self, value: self._set_object_list("_imagesequences", value),
        doc="list of ImageSequences contained in this group",
    )

    regionsofinterest = property(
        fget=lambda self: self._get_object_list("_regionsofinterest"),
        fset=lambda self, value: self._set_object_list("_regionsofinterest", value),
        doc="list of RegionOfInterest objects contained in this group",
    )

    spiketrains = property(
        fget=lambda self: self._get_object_list("_spiketrains"),
        fset=lambda self, value: self._set_object_list("_spiketrains", value),
        doc="list of SpikeTrains contained in this group",
    )

    segments = property(
        fget=lambda self: self._get_object_list("_segments"),
        fset=lambda self, value: self._set_object_list("_segments", value),
        doc="list of Segments contained in this group",
    )

    groups = property(
        fget=lambda self: self._get_object_list("_groups"),
        fset=lambda self, value: self._set_object_list("_groups", value),
        doc="list of Groups contained in this group",
    )

    def add(self, *objects):
        """Add a new Neo object to the Group"""
        for obj in objects:
            if self.allowed_types and not isinstance(obj, self.allowed_types):
                raise TypeError(f"This Group can only contain {self.allowed_types}, " f"but not {type(obj)}")
        super().add(*objects)

    def walk(self):
        """
        Walk the tree of subgroups
        """
        yield self
        for grp in self.groups:
            yield from grp.walk()
